#
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2007-2009  Brian G. Matherly
# Copyright (C) 2008       Raphael Ackermann
#               2002-2003  Donald A. Peterson
#               2003       Alex Roitman
#               2009       Benny Malengier
#               2010       Peter Landgren
# Copyright (C) 2011       Adam Stein <adam@csh.rit.edu>
#               2011-2012  Harald Rosemann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""LaTeX document generator"""

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
from bisect import bisect
import re
import os
import logging

try:
    from PIL import Image

    HAVE_PIL = True
except ImportError:
    HAVE_PIL = False

# ----------------------------------------------------------------------- -
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug.docgen import (
    BaseDoc,
    TextDoc,
    PAPER_LANDSCAPE,
    FONT_SANS_SERIF,
    URL_PATTERN,
)
from gramps.gen.plug.docbackend import DocBackend
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

_LOG = logging.getLogger(".latexdoc")

_CLICKABLE = "\\url{\\1}"

# ------------------------------------------------------------------------
#
# Special settings for LaTeX output
#
# ------------------------------------------------------------------------
#   For an interim mark e.g. for an intended linebreak I use a special pattern.
#   It shouldn't interfere with normal text. In LaTeX character '&' is used
#   for column separation in tables and may occur there in series. The pattern
#   is used here before column separation is set. On the other hand incoming
#   text can't show this pattern for it would have been replaced by '\&\&'.
#   So the choosen pattern will do the job without confusion:

SEPARATION_PAT = "&&"

# ------------------------------------------------------------------------
#
# LaTeX Article Template
#
# ------------------------------------------------------------------------

_LATEX_TEMPLATE_1 = "\\documentclass[%s]{article}\n"
_LATEX_TEMPLATE = """%
%
\\usepackage[T1]{fontenc}%
%
% We use latin1 encoding at a minimum by default.
% Gramps uses unicode UTF-8 encoding for its
% international support. LaTeX can deal gracefully
% with unicode encoding by using the ucs style invoked
% when utf8 is specified as an option to the inputenc
% package. This package is included by default in some
% installations, but not in others, so we do not make it
% the default.  Uncomment the first line if you wish to use it
% (If you do not have ucs.sty, you may obtain it from
%  http://www.tug.org/tex-archive/macros/latex/contrib/supported/unicode/)
%
%\\usepackage[latin1]{inputenc}%
\\usepackage[latin1,utf8]{inputenc}%
\\usepackage{graphicx}% Extended graphics support
\\usepackage{longtable}% For multi-page tables
\\usepackage{calc}% For some calculations
\\usepackage{ifthen}% For table width calculations
\\usepackage{ragged2e}% For left aligning with hyphenation
\\usepackage{wrapfig}% wrap pictures in text
\\usepackage[normalem]{ulem}% For strikeout
%
% Depending on your LaTeX installation, the margins may be too
% narrow.  This can be corrected by uncommenting the following
% two lines and adjusting the width appropriately. The example
% removes 0.5in from each margin. (Adds 1 inch to the text)
%\\addtolength{\\oddsidemargin}{-0.5in}%
%\\addtolength{\\textwidth}{1.0in}%
%
% Vertical spacing between paragraphs:
% take one of three possibilities or modify to your taste:
%\\setlength{\\parskip}{1.0ex plus0.2ex minus0.2ex}%
\\setlength{\\parskip}{1.5ex plus0.3ex minus0.3ex}%
%\\setlength{\\parskip}{2.0ex plus0.4ex minus0.4ex}%
%
% Vertical spacing between lines:
% take one of three possibilities or modify to your taste:
\\renewcommand{\\baselinestretch}{1.0}%
%\\renewcommand{\\baselinestretch}{1.1}%
%\\renewcommand{\\baselinestretch}{1.2}%
%
% Indentation; substitute for '1cm' of gramps, 2.5em is right for 12pt
% take one of three possibilities or modify to your taste:
\\newlength{\\grbaseindent}%
%\\setlength{\\grbaseindent}{3.0em}%
\\setlength{\\grbaseindent}{2.5em}%
%\\setlength{\\grbaseindent}{2.0em}%
%
%
% -------------------------------------------------------------
% New lengths, counters and commands for calculations in tables
% -------------------------------------------------------------
%
\\newlength{\\grtabwidth}%
\\newlength{\\grtabprepos}%
\\newlength{\\grreqwidth}%
\\newlength{\\grtempwd}%
\\newlength{\\grmaxwidth}%
\\newlength{\\grprorated}%
\\newlength{\\grxwd}%
\\newlength{\\grwidthused}%
\\newlength{\\grreduce}%
\\newlength{\\grcurcolend}%
\\newlength{\\grspanwidth}%
\\newlength{\\grleadlabelwidth}%
\\newlength{\\grminpgindent}%
\\newlength{\\grlistbacksp}%
\\newlength{\\grpictsize}%
\\newlength{\\grmaxpictsize}%
\\newlength{\\grtextsize}%
\\newlength{\\grmaxtextsize}%
\\newcounter{grtofixcnt}%
\\newcounter{grxwdcolcnt}%
%
%
\\newcommand{\\grinitlength}[2]{%
  \\ifthenelse{\\isundefined{#1}}%
    {\\newlength{#1}}{}%
  \\setlength{#1}{#2}%
}%
%
\\newcommand{\\grinittab}[2]{%    #1: tabwidth, #2 = 1.0/anz-cols
  \\setlength{\\grtabwidth}{#1}%
  \\setlength{\\grprorated}{#2\\grtabwidth}%
  \\setlength{\\grwidthused}{0em}%
  \\setlength{\\grreqwidth}{0em}%
  \\setlength{\\grmaxwidth }{0em}%
  \\setlength{\\grxwd}{0em}%
  \\setlength{\\grtempwd}{0em}%
  \\setlength{\\grpictsize}{0em}%
  \\setlength{\\grmaxpictsize}{0em}%
  \\setlength{\\grtextsize}{0em}%
  \\setlength{\\grmaxtextsize}{0em}%
  \\setlength{\\grcurcolend}{0em}%
  \\setcounter{grxwdcolcnt}{0}%
  \\setcounter{grtofixcnt}{0}%  number of wide cols%
  \\grinitlength{\\grcolbega}{0em}% beg of first col
}%
%
\\newcommand{\\grmaxvaltofirst}[2]{%
  \\ifthenelse{\\lengthtest{#1 < #2}}%
    {\\setlength{#1}{#2}}{}%
}%
%
\\newcommand{\\grsetreqfull}{%
  \\grmaxvaltofirst{\\grmaxpictsize}{\\grpictsize}%
  \\grmaxvaltofirst{\\grmaxtextsize}{\\grtextsize}%
}%
%
\\newcommand{\\grsetreqpart}[1]{%
  \\addtolength{\\grtextsize}{#1 - \\grcurcolend}%
  \\addtolength{\\grpictsize}{#1 - \\grcurcolend}%
  \\grsetreqfull%
}%
%
\\newcommand{\\grdividelength}{%
 \\setlength{\\grtempwd}{\\grtabwidth - \\grwidthused}%
%    rough division of lengths:
%    if 0 < #1 <= 10: \\grxwd = ~\\grtempwd / grtofixcnt
%    otherwise: \\grxwd =  \\grprorated
 \\ifthenelse{\\value{grtofixcnt} > 0}%
  {\\ifthenelse{\\value{grtofixcnt}=1}%
                    {\\setlength{\\grxwd}{\\grtempwd}}{%
    \\ifthenelse{\\value{grtofixcnt}=2}
                    {\\setlength{\\grxwd}{0.5\\grtempwd}}{%
     \\ifthenelse{\\value{grtofixcnt}=3}
                    {\\setlength{\\grxwd}{0.333\\grtempwd}}{%
      \\ifthenelse{\\value{grtofixcnt}=4}
                    {\\setlength{\\grxwd}{0.25\\grtempwd}}{%
       \\ifthenelse{\\value{grtofixcnt}=5}
                    {\\setlength{\\grxwd}{0.2\\grtempwd}}{%
        \\ifthenelse{\\value{grtofixcnt}=6}
                    {\\setlength{\\grxwd}{0.166\\grtempwd}}{%
         \\ifthenelse{\\value{grtofixcnt}=7}
                    {\\setlength{\\grxwd}{0.143\\grtempwd}}{%
          \\ifthenelse{\\value{grtofixcnt}=8}
                    {\\setlength{\\grxwd}{0.125\\grtempwd}}{%
           \\ifthenelse{\\value{grtofixcnt}=9}
                    {\\setlength{\\grxwd}{0.111\\grtempwd}}{%
            \\ifthenelse{\\value{grtofixcnt}=10}
                    {\\setlength{\\grxwd}{0.1\\grtempwd}}{%
             \\setlength{\\grxwd}{\\grprorated}% give up, take \\grprorated%
    }}}}}}}}}}%
  \\setlength{\\grreduce}{0em}%
  }{\\setlength{\\grxwd}{0em}}%
}%
%
\\newcommand{\\grtextneedwidth}[1]{%
  \\settowidth{\\grtempwd}{#1}%
  \\grmaxvaltofirst{\\grtextsize}{\\grtempwd}%
}%
%
\\newcommand{\\grcolsfirstfix}[5]{%
  \\grinitlength{#1}{\\grcurcolend}%
  \\grinitlength{#3}{0em}%
  \\grinitlength{#4}{\\grmaxpictsize}%
  \\grinitlength{#5}{\\grmaxtextsize}%
  \\grinitlength{#2}{#5}%
  \\grmaxvaltofirst{#2}{#4}%
  \\addtolength{#2}{2\\tabcolsep}%
  \\grmaxvaltofirst{\\grmaxwidth}{#2}%
  \\ifthenelse{\\lengthtest{#2 < #4} \\or \\lengthtest{#2 < \\grprorated}}%
    { \\setlength{#3}{#2}%
      \\addtolength{\\grwidthused}{#2} }%
    { \\stepcounter{grtofixcnt} }%
  \\addtolength{\\grcurcolend}{#2}%
}%
%
\\newcommand{\\grcolssecondfix}[4]{%
  \\ifthenelse{\\lengthtest{\\grcurcolend < \\grtabwidth}}%
    { \\setlength{#3}{#2} }%
    { \\addtolength{#1}{-\\grreduce}%
      \\ifthenelse{\\lengthtest{#2 = \\grmaxwidth}}%
        { \\stepcounter{grxwdcolcnt}}%
        { \\ifthenelse{\\lengthtest{#3 = 0em} \\and %
                       \\lengthtest{#4 > 0em}}%
            { \\setlength{\\grtempwd}{#4}%
              \\grmaxvaltofirst{\\grtempwd}{\\grxwd}%
              \\addtolength{\\grreduce}{#2 - \\grtempwd}%
              \\setlength{#2}{\\grtempwd}%
              \\addtolength{\\grwidthused}{#2}%
              \\addtocounter{grtofixcnt}{-1}%
              \\setlength{#3}{#2}%
            }{}%
        }%
    }%
}%
%
\\newcommand{\\grcolsthirdfix}[3]{%
  \\ifthenelse{\\lengthtest{\\grcurcolend < \\grtabwidth}}%
    {}{ \\addtolength{#1}{-\\grreduce}%
        \\ifthenelse{\\lengthtest{#3 = 0em} \\and %
                     \\lengthtest{#2 < \\grmaxwidth}}%
          { \\ifthenelse{\\lengthtest{#2 < 0.5\\grmaxwidth}}%
              { \\setlength{\\grtempwd}{0.5\\grxwd}%
                \\grmaxvaltofirst{\\grtempwd}{0.7\\grprorated}}%
              { \\setlength{\\grtempwd}{\\grxwd}}%
            \\addtolength{\\grreduce}{#2 - \\grtempwd}%
            \\setlength{#2}{\\grtempwd}%
            \\addtolength{\\grwidthused}{#2}%
            \\addtocounter{grtofixcnt}{-1}%
            \\setlength{#3}{#2}%
          }{}%
      }%
}%
%
\\newcommand{\\grcolsfourthfix}[3]{%
  \\ifthenelse{\\lengthtest{\\grcurcolend < \\grtabwidth}}%
    {}{ \\addtolength{#1}{-\\grreduce}%
        \\ifthenelse{\\lengthtest{#3 = 0em}}%
          { \\addtolength{\\grreduce}{#2 - \\grxwd}%
            \\setlength{#2}{\\grxwd}%
            \\setlength{#3}{#2}%
          }{}%
      }%
}%
%
\\newcommand{\\grgetspanwidth}[4]{%
  \\grinitlength{#1}{#3 - #2 + #4}%
}%
%
\\newcommand{\\tabheadstrutceil}{%
  \\rule[0.0ex]{0.00em}{3.5ex}}%
\\newcommand{\\tabheadstrutfloor}{%
  \\rule[-2.0ex]{0.00em}{2.5ex}}%
\\newcommand{\\tabrowstrutceil}{%
  \\rule[0.0ex]{0.00em}{2.9ex}}%
\\newcommand{\\tabrowstrutfloor}{%
  \\rule[-0.1ex]{0.00em}{2.0ex}}%
%
\\newcommand{\\grempty}[1]{}%
%
\\newcommand{\\graddvdots}[1]{%
  \\hspace*{\\fill}\\hspace*{\\fill}\\raisebox{#1}{\\vdots}%
}%
%
\\newcommand{\\grtabpgbreak}[4]{%
  #1 { \\parbox[t]{ #2 - 2\\tabcolsep}{\\tabheadstrutceil\\hspace*{\\fill}%
  \\raisebox{#4}{\\vdots} #3{#4} \\hspace*{\\fill}\\tabheadstrutfloor}}%
}%
%
\\newcommand{\\grcolpart}[3]{%
  #1 { \\parbox[t]{ #2 - 2\\tabcolsep}%
  {\\tabrowstrutceil #3~\\\\[-1.6ex]\\tabrowstrutfloor}}%
}%
%
\\newcommand{\\grminpghead}[2]{%
  \\setlength{\\grminpgindent}{#1\\grbaseindent-\\grlistbacksp}%
  \\hspace*{\\grminpgindent}%
  \\ifthenelse{\\not \\lengthtest{#2em > 0em}}%
    {\\begin{minipage}[t]{\\textwidth -\\grminpgindent}}%
    {\\begin{minipage}[t]{\\textwidth -\\grminpgindent%
        -#2\\grbaseindent -4\\tabcolsep}}%
}%
%
\\newcommand{\\grminpgtail}{%
  \\end{minipage}\\parindent0em%
}%
%
\\newcommand{\\grlisthead}[1]{%
  \\begin{list}{#1}%
    { \\setlength{\\labelsep}{0.5em}%
      \\setlength{\\labelwidth}{\\grleadlabelwidth}%
      \\setlength{\\leftmargin}{\\grlistbacksp}%
    }\\item%
}%
%
\\newcommand{\\grlisttail}{%
  \\end{list}%
}%
%
\\newcommand{\\grprepleader}[1]{%
  \\settowidth{\\grtempwd}{#1}%
  \\ifthenelse{\\lengthtest{\\grtempwd > \\grleadlabelwidth}}%
    { \\setlength{\\grleadlabelwidth}{\\grtempwd}}{}%
  \\setlength{\\grlistbacksp}{\\grleadlabelwidth + 1.0em}%
}%
%
\\newcommand{\\grprepnoleader}{%
  \\setlength{\\grleadlabelwidth}{0em}%
  \\setlength{\\grlistbacksp}{0em}%
}%
%
\\newcommand{\\grmkpicture}[4]{%
    \\begin{wrapfigure}{r}{#2\\grbaseindent}%
      \\vspace{-6ex}%
      \\begin{center}%
      \\includegraphics[%
        width= #2\\grbaseindent,%
        height= #3\\grbaseindent,%
          keepaspectratio]%
        {#1}\\\\%
      {\\RaggedRight\\footnotesize#4}%
      \\end{center}%
    \\end{wrapfigure}%
    \\settowidth{\\grtempwd}{\\footnotesize#4}%
    \\setlength{\\grxwd}{#2\\grbaseindent}%
    \\ifthenelse{\\lengthtest{\\grtempwd < 0.7\\grxwd}}%
                    {\\setlength{\\grxwd}{1ex}}{%
      \\ifthenelse{\\lengthtest{\\grtempwd < 1.2\\grxwd}}%
                    {\\setlength{\\grxwd}{2ex}}{%
        \\ifthenelse{\\lengthtest{\\grtempwd < 1.8\\grxwd}}%
                    {\\setlength{\\grxwd}{6ex}}{%
          \\ifthenelse{\\lengthtest{\\grtempwd < 2.0\\grxwd}}%
                    {\\setlength{\\grxwd}{10ex}}{%
                     \\setlength{\\grxwd}{12ex}}%
                    }}}%
  \\setlength{\\grtempwd}{#3\\grbaseindent + \\grxwd}%
  \\rule[-\\grtempwd]{0pt}{\\grtempwd}%
  \\setlength{\\grtabprepos}{-\\grtempwd}%
}%
%
%
\\begin{document}%
"""


# ------------------------------------------------------------------------
#
# Font size table and function
#
# ------------------------------------------------------------------------

# These tables correlate font sizes to LaTeX.  The first table contains
# typical font sizes in points.  The second table contains the standard
# LaTeX font size names. Since we use bisect to map the first table to the
# second, we are guaranteed that any font less than 6 points is 'tiny', fonts
# from 6-7 points are 'script', etc. and fonts greater than or equal to 22
# are considered 'Huge'.  Note that fonts from 12-13 points are not given a
# LaTeX font size name but are considered "normal."

_FONT_SIZES = [6, 8, 10, 12, 14, 16, 18, 20, 22]
_FONT_NAMES = [
    "tiny",
    "scriptsize",
    "footnotesize",
    "small",
    "",
    "large",
    "Large",
    "LARGE",
    "huge",
    "Huge",
]


def map_font_size(fontsize):
    """Map font size in points to LaTeX font size"""
    return _FONT_NAMES[bisect(_FONT_SIZES, fontsize)]


# ------------------------------------------------------------------------
#
# auxiliaries to facilitate table construction
#
# ------------------------------------------------------------------------

# patterns for regular expressions, module re:
TBLFMT_PAT = re.compile(r"({\|?)l(\|?})")

# constants for routing in table construction:
(CELL_BEG, CELL_TEXT, CELL_END, ROW_BEG, ROW_END, TAB_BEG, TAB_END) = list(range(7))
FIRST_ROW, SUBSEQ_ROW = list(range(2))


def get_charform(col_num):
    """
    Transfer column number to column charakter,
    limited to letters within a-z;
    26, there is no need for more.
    early test of column count in start_table()
    """
    if col_num > ord("z") - ord("a"):
        raise ValueError(
            "".join(
                (
                    "\n number of table columns is ",
                    repr(col_num),
                    "\n                     should be <= ",
                    repr(ord("z") - ord("a")),
                )
            )
        )
    return chr(ord("a") + col_num)


def get_numform(col_char):
    return ord(col_char) - ord("a")


# ------------------------------------------
#   row_alph_counter = str_incr(MULTCOL_COUNT_BASE)
#
#   'aaa' is sufficient for up to 17576 multicolumns in each table;
#   do you need more?
#   uncomment one of the two lines
MULTCOL_COUNT_BASE = "aaa"
# MULTCOL_COUNT_BASE = 'aaaa'
# ------------------------------------------


def str_incr(str_counter):
    """for counting table rows"""
    lili = list(str_counter)
    while 1:
        yield "".join(lili)
        if "".join(lili) == len(lili) * "z":
            raise ValueError(
                "".join(
                    (
                        "\n can't increment string ",
                        "".join(lili),
                        " of length ",
                        str(len(lili)),
                    )
                )
            )
        for i in reversed(lili):
            if lili[i] < "z":
                lili[i] = chr(ord(lili[i]) + 1)
                break
            else:
                lili[i] = "a"


# ------------------------------------------------------------------------
#
# Structure of Table-Memory
#
# ------------------------------------------------------------------------


class TabCell:
    def __init__(self, colchar, span, head, content):
        self.colchar = colchar
        self.span = span
        self.head = head
        self.content = content


class TabRow:
    def __init__(self):
        self.cells = []
        self.tail = ""
        self.addit = ""  # for: \\hline, \\cline{}


class TabMem:
    def __init__(self, head):
        self.head = head
        self.tail = ""
        self.rows = []


# ------------------------------------------------------------------------
#
# Functions for docbackend
#
# ------------------------------------------------------------------------
def latexescape(text):
    """
    Escape the following special characters: & $ % # _ { }
    """
    text = text.replace("&", "\\&")
    text = text.replace("$", "\\$")
    text = text.replace("%", "\\%")
    text = text.replace("#", "\\#")
    text = text.replace("_", "\\_")
    text = text.replace("{", "\\{")
    text = text.replace("}", "\\}")
    # replace character unknown to LaTeX
    text = text.replace("→", "$\\longrightarrow$")
    return text


def latexescapeverbatim(text):
    """
    Escape special characters and also make sure that LaTeX respects whitespace
    and newlines correctly.
    """
    text = latexescape(text)
    text = text.replace(" ", "\\ ")
    text = text.replace("\n", "~\\newline \n")
    # spaces at begin are normally ignored, make sure they are not.
    # due to above a space at begin is now \newline\n\
    text = text.replace("\\newline\n\\ ", "\\newline\n\\hspace*{0.1\\grbaseindent}\\ ")
    return text


# ------------------------------------------------------------------------
#
# Document Backend class for cairo docs
#
# ------------------------------------------------------------------------


class LaTeXBackend(DocBackend):
    """
    Implementation of docbackend for latex docs.
    File and File format management for latex docs
    """

    # overwrite base class attributes, they become static var of LaTeXDoc
    SUPPORTED_MARKUP = [
        DocBackend.BOLD,
        DocBackend.ITALIC,
        DocBackend.UNDERLINE,
        DocBackend.FONTSIZE,
        DocBackend.FONTFACE,
        DocBackend.SUPERSCRIPT,
        DocBackend.SUBSCRIPT,
        DocBackend.STRIKETHROUGH,
    ]

    STYLETAG_MARKUP = {
        DocBackend.BOLD: ("\\textbf{", "}"),
        DocBackend.ITALIC: ("\\textit{", "}"),
        DocBackend.UNDERLINE: ("\\underline{", "}"),
        DocBackend.SUPERSCRIPT: ("\\textsuperscript{", "}"),
        DocBackend.SUBSCRIPT: ("\\textsubscript{", "}"),
        DocBackend.STRIKETHROUGH: ("\\sout{", "}"),
    }

    ESCAPE_FUNC = lambda x: latexescape

    def setescape(self, preformatted=False):
        """
        LaTeX needs two different escape functions depending on the type.
        This function allows to switch the escape function
        """
        if not preformatted:
            LaTeXBackend.ESCAPE_FUNC = lambda x: latexescape
        else:
            LaTeXBackend.ESCAPE_FUNC = lambda x: latexescapeverbatim

    def _create_xmltag(self, type, value):
        r"""
        overwrites the method in DocBackend.
        creates the latex tags needed for non bool style types we support:
            FONTSIZE : use different \large denomination based
                                        on size
                                     : very basic, in mono in the font face
                                        then we use {\ttfamily }
        """
        if type not in self.SUPPORTED_MARKUP:
            return None
        elif type == DocBackend.FONTSIZE:
            # translate size in point to something LaTeX can work with
            fontsize = map_font_size(value)
            if fontsize:
                return ("{\\" + fontsize + " ", "}")
            else:
                return ("", "")

        elif type == DocBackend.FONTFACE:
            if "MONO" in value.upper():
                return ("{\\ttfamily ", "}")
            elif "ROMAN" in value.upper():
                return ("{\\rmfamily ", "}")
        return None

    def _checkfilename(self):
        """
        Check to make sure filename satisfies the standards for this filetype
        """
        if not self._filename.endswith(".tex"):
            self._filename = self._filename + ".tex"


# ------------------------------------------------------------------------
#
# Paragraph Handling
#
# ------------------------------------------------------------------------


class TexFont:
    def __init__(self, style=None):
        if style:
            self.font_beg = style.font_beg
            self.font_end = style.font_end
            self.left_indent = style.left_indent
            self.first_line_indent = style.first_line_indent
        else:
            self.font_beg = ""
            self.font_end = ""
            self.left_indent = ""
            self.first_line_indent = ""


# ------------------------------------------------------------------
#
# LaTeXDoc
#
# ------------------------------------------------------------------


class LaTeXDoc(BaseDoc, TextDoc):
    """LaTeX document interface class. Derived from BaseDoc"""

    #   ---------------------------------------------------------------
    #   some additional variables
    #   ---------------------------------------------------------------
    in_table = False
    in_multrow_cell = False  #   for tab-strukt: cols of rows
    pict = ""
    pict_in_table = False
    pict_width = 0
    pict_height = 0
    textmem = []
    in_title = True

    #   ---------------------------------------------------------------
    #   begin of table special treatment
    #   ---------------------------------------------------------------
    def emit(self, text, tab_state=CELL_TEXT, span=1):
        """
        Hand over all text but tables to self._backend.write(), (line 1-2).
        In case of tables pass to specal treatment below.
        """
        if not self.in_table:  # all stuff but table
            self._backend.write(text)
        else:
            self.handle_table(text, tab_state, span)

    def handle_table(self, text, tab_state, span):
        """
        Collect tables elements in an adequate cell/row/table structure and
        call for LaTeX width calculations and writing out
        """
        if tab_state == CELL_BEG:
            # here text is head
            self.textmem = []
            self.curcol_char = get_charform(self.curcol - 1)
            if span > 1:  # phantom columns prior to multicolumns
                for col in range(self.curcol - span, self.curcol - 1):
                    col_char = get_charform(col)
                    phantom = TabCell(col_char, 0, "", "")
                    self.tabrow.cells.append(phantom)
            self.tabcell = TabCell(self.curcol_char, span, text, "")
        elif tab_state == CELL_TEXT:
            self.textmem.append(text)
        elif tab_state == CELL_END:  # text == ''
            self.tabcell.content = "".join(self.textmem).strip()

            if self.tabcell.content.find("\\centering") != -1:
                self.tabcell.content = self.tabcell.content.replace("\\centering", "")
                self.tabcell.head = re.sub(TBLFMT_PAT, "\\1c\\2", self.tabcell.head)
            self.tabrow.cells.append(self.tabcell)
            self.textmem = []
        elif tab_state == ROW_BEG:
            self.tabrow = TabRow()
        elif tab_state == ROW_END:
            self.tabrow.addit = text  # text: \\hline, \\cline{}
            self.tabrow.tail = "".join(self.textmem)  # \\\\ row-termination
            if self.in_multrow_cell:  #   cols of rows: convert to rows of cols
                self.repack_row()
            else:
                self.tabmem.rows.append(self.tabrow)
        elif tab_state == TAB_BEG:  # text: \\begin{longtable}[l]{
            self._backend.write(
                "".join(("\\grinittab{\\textwidth}{", repr(1.0 / self.numcols), "}%\n"))
            )
            self.tabmem = TabMem(text)
        elif tab_state == TAB_END:  # text: \\end{longtable}
            self.tabmem.tail = text

            # table completed, calc widths and write out
            self.calc_latex_widths()
            self.write_table()

    def repack_row(self):
        """
        Transpose contents contained in a row of cols of cells
        to rows of cells with corresponding contents.
        Cols of the mult-row-cell are ended by SEPARATION_PAT
        """
        # if last col empty: delete
        if self.tabrow.cells[-1].content == "":
            del self.tabrow.cells[-1]
            self.numcols -= 1

        # extract cell.contents
        bare_contents = [
            cell.content.strip(SEPARATION_PAT).replace("\n", "").split(SEPARATION_PAT)
            for cell in self.tabrow.cells
        ]

        # mk equal length & transpose
        num_new_rows = max([len(mult_row_cont) for mult_row_cont in bare_contents])
        cols_equ_len = []
        for mrc in bare_contents:
            for i in range(num_new_rows - len(mrc)):
                mrc.append("")
            cols_equ_len.append(mrc)
        transp_cont = list(zip(*cols_equ_len))

        # picts? extract
        first_cell, last_cell = (0, self.numcols)
        if self.pict_in_table:
            if transp_cont[0][-1].startswith("\\grmkpicture"):
                self.pict = transp_cont[0][-1]
                last_cell -= 1
                self.numcols -= 1
                self._backend.write(
                    "".join(
                        (
                            "\\addtolength{\\grtabwidth}{-",
                            repr(self.pict_width),
                            "\\grbaseindent -2\\tabcolsep}%\n",
                        )
                    )
                )
            self.pict_in_table = False

        # new row-col structure
        for row in range(num_new_rows):
            new_row = TabRow()
            for i in range(first_cell, last_cell):
                new_cell = TabCell(
                    get_charform(i + first_cell),
                    self.tabrow.cells[i].span,
                    self.tabrow.cells[i].head,
                    transp_cont[row][i + first_cell],
                )
                new_row.cells.append(new_cell)
            new_row.tail = self.tabrow.tail
            new_row.addit = ""
            self.tabmem.rows.append(new_row)

        self.tabmem.rows[-1].addit = self.tabrow.addit
        self.in_multrow_cell = False

    def calc_latex_widths(self):
        """
        Control width settings in latex table construction
        Evaluations are set up here and passed to LaTeX
        to calculate required and to fix suitable widths.
        ??? Can all this be done exclusively in TeX? Don't know how.
        """
        tabcol_chars = []
        for col_num in range(self.numcols):
            col_char = get_charform(col_num)
            tabcol_chars.append(col_char)
            for row in self.tabmem.rows:
                cell = row.cells[col_num]
                if cell.span == 0:
                    continue
                if cell.content.startswith("\\grmkpicture"):
                    self._backend.write(
                        "".join(
                            (
                                "\\setlength{\\grpictsize}{",
                                self.pict_width,
                                "\\grbaseindent}%\n",
                            )
                        )
                    )
                else:
                    for part in cell.content.split(SEPARATION_PAT):
                        self._backend.write(
                            "".join(("\\grtextneedwidth{", part, "}%\n"))
                        )
                    row.cells[col_num].content = cell.content.replace(
                        SEPARATION_PAT, "~\\newline \n"
                    )

                if cell.span == 1:
                    self._backend.write("".join(("\\grsetreqfull%\n")))
                elif cell.span > 1:
                    self._backend.write(
                        "".join(
                            (
                                "\\grsetreqpart{\\grcolbeg",
                                get_charform(get_numform(cell.colchar) - cell.span + 1),
                                "}%\n",
                            )
                        )
                    )

            self._backend.write(
                "".join(
                    (
                        "\\grcolsfirstfix",
                        " {\\grcolbeg",
                        col_char,
                        "}{\\grtempwidth",
                        col_char,
                        "}{\\grfinalwidth",
                        col_char,
                        "}{\\grpictreq",
                        col_char,
                        "}{\\grtextreq",
                        col_char,
                        "}%\n",
                    )
                )
            )

        self._backend.write("".join(("\\grdividelength%\n")))
        for col_char in tabcol_chars:
            self._backend.write(
                "".join(
                    (
                        "\\grcolssecondfix",
                        " {\\grcolbeg",
                        col_char,
                        "}{\\grtempwidth",
                        col_char,
                        "}{\\grfinalwidth",
                        col_char,
                        "}{\\grpictreq",
                        col_char,
                        "}%\n",
                    )
                )
            )

        self._backend.write("".join(("\\grdividelength%\n")))
        for col_char in tabcol_chars:
            self._backend.write(
                "".join(
                    (
                        "\\grcolsthirdfix",
                        " {\\grcolbeg",
                        col_char,
                        "}{\\grtempwidth",
                        col_char,
                        "}{\\grfinalwidth",
                        col_char,
                        "}%\n",
                    )
                )
            )

        self._backend.write("".join(("\\grdividelength%\n")))
        for col_char in tabcol_chars:
            self._backend.write(
                "".join(
                    (
                        "\\grcolsfourthfix",
                        " {\\grcolbeg",
                        col_char,
                        "}{\\grtempwidth",
                        col_char,
                        "}{\\grfinalwidth",
                        col_char,
                        "}%\n",
                    )
                )
            )

        self.multcol_alph_counter = str_incr(MULTCOL_COUNT_BASE)
        for row in self.tabmem.rows:
            for cell in row.cells:
                if cell.span > 1:
                    multcol_alph_id = next(self.multcol_alph_counter)
                    self._backend.write(
                        "".join(
                            (
                                "\\grgetspanwidth{",
                                "\\grspanwidth",
                                multcol_alph_id,
                                "}{\\grcolbeg",
                                get_charform(get_numform(cell.colchar) - cell.span + 1),
                                "}{\\grcolbeg",
                                cell.colchar,
                                "}{\\grtempwidth",
                                cell.colchar,
                                "}%\n",
                            )
                        )
                    )

    def write_table(self):
        # Choosing RaggedRight (with hyphenation) in table and
        # provide manually adjusting of column widths
        self._backend.write(
            "".join(
                (
                    "%\n",
                    self.pict,
                    "%\n%\n",
                    "%  ==> Comment out one of the two lines ",
                    'by a leading "%" (first position)\n',
                    "{ \\RaggedRight%      left align with hyphenation in table \n",
                    "%{%                no left align in table \n%\n",
                    "%  ==>  You may add pos or neg values ",
                    "to the following ",
                    repr(self.numcols),
                    " column widths %\n",
                )
            )
        )
        for col_num in range(self.numcols):
            self._backend.write(
                "".join(
                    (
                        "\\addtolength{\\grtempwidth",
                        get_charform(col_num),
                        "}{+0.0cm}%\n",
                    )
                )
            )
        self._backend.write("%  === %\n")

        # adjust & open table':
        if self.pict:
            self._backend.write(
                "".join(
                    (
                        "%\n\\vspace{\\grtabprepos}%\n",
                        "\\setlength{\\grtabprepos}{0ex}%\n",
                    )
                )
            )
            self.pict = ""
        self._backend.write("".join(self.tabmem.head))

        # special treatment at begin of longtable for heading and
        # closing at top and bottom of table
        # and parts of it at pagebreak separating
        self.multcol_alph_counter = str_incr(MULTCOL_COUNT_BASE)
        splitting_row = self.mk_splitting_row(self.tabmem.rows[FIRST_ROW])
        self.multcol_alph_counter = str_incr(MULTCOL_COUNT_BASE)
        complete_row = self.mk_complete_row(self.tabmem.rows[FIRST_ROW])

        self._backend.write(splitting_row)
        self._backend.write("\\endhead%\n")
        self._backend.write(splitting_row.replace("{+2ex}", "{-2ex}"))
        self._backend.write("\\endfoot%\n")
        if self.head_line:
            self._backend.write("\\hline%\n")
            self.head_line = False
        else:
            self._backend.write("%\n")
        self._backend.write(complete_row)
        self._backend.write("\\endfirsthead%\n")
        self._backend.write("\\endlastfoot%\n")

        # hand over subsequent rows
        for row in self.tabmem.rows[SUBSEQ_ROW:]:
            self._backend.write(self.mk_complete_row(row))

        # close table by '\\end{longtable}', end '{\\RaggedRight' or '{' by '}'
        self._backend.write("".join(("".join(self.tabmem.tail), "}%\n\n")))

    def mk_splitting_row(self, row):
        splitting = []
        add_vdots = "\\grempty"
        for cell in row.cells:
            if cell.span == 0:
                continue
            if not splitting and get_numform(cell.colchar) == self.numcols - 1:
                add_vdots = "\\graddvdots"
            if cell.span == 1:
                cell_width = "".join(("\\grtempwidth", cell.colchar))
            else:
                cell_width = "".join(("\\grspanwidth", next(self.multcol_alph_counter)))
            splitting.append(
                "".join(
                    (
                        "\\grtabpgbreak{",
                        cell.head,
                        "}{",
                        cell_width,
                        "}{",
                        add_vdots,
                        "}{+2ex}%\n",
                    )
                )
            )
        return "".join((" & ".join(splitting), "%\n", row.tail))

    def mk_complete_row(self, row):
        complete = []
        for cell in row.cells:
            if cell.span == 0:
                continue
            elif cell.span == 1:
                cell_width = "".join(("\\grtempwidth", cell.colchar))
            else:
                cell_width = "".join(("\\grspanwidth", next(self.multcol_alph_counter)))
            complete.append(
                "".join(
                    (
                        "\\grcolpart{%\n  ",
                        cell.head,
                        "}{%\n",
                        cell_width,
                        "}{%\n  ",
                        cell.content,
                        "%\n}%\n",
                    )
                )
            )
        return "".join((" & ".join(complete), "%\n", row.tail, row.addit))

    #       ---------------------------------------------------------------------
    #       end of special table treatment
    #       ---------------------------------------------------------------------

    def page_break(self):
        "Forces a page break, creating a new page"
        self.emit("\\newpage%\n")

    def open(self, filename):
        """Opens the specified file, making sure that it has the
        extension of .tex"""
        self._backend = LaTeXBackend(filename)
        self._backend.open()

        # Font size control seems to be limited. For now, ignore
        # any style constraints, and use 12pt as the default

        options = "12pt"

        if self.paper.get_orientation() == PAPER_LANDSCAPE:
            options = options + ",landscape"

        # Paper selections are somewhat limited on a stock installation.
        # If the user picks something not listed here, we'll just accept
        # the default of the user's LaTeX installation (usually letter).
        paper_name = self.paper.get_size().get_name().lower()
        if paper_name in ["a4", "a5", "legal", "letter"]:
            options += "," + paper_name + "paper"

        # Use the article template, T1 font encodings, and specify
        # that we should use Latin1 and unicode character encodings.
        self.emit(_LATEX_TEMPLATE_1 % options)
        self.emit(_LATEX_TEMPLATE)

        self.in_list = False
        self.in_table = False
        self.head_line = False

        # Establish some local styles for the report
        self.latexstyle = {}
        self.latex_font = {}

        style_sheet = self.get_style_sheet()
        for style_name in style_sheet.get_paragraph_style_names():
            style = style_sheet.get_paragraph_style(style_name)
            font = style.get_font()
            size = font.get_size()

            self.latex_font[style_name] = TexFont()
            thisstyle = self.latex_font[style_name]

            thisstyle.font_beg = ""
            thisstyle.font_end = ""
            # Is there special alignment?  (default is left)
            align = style.get_alignment_text()
            if align == "center":
                thisstyle.font_beg += "{\\centering"
                thisstyle.font_end = "".join(("\n\n}", thisstyle.font_end))
            elif align == "right":
                thisstyle.font_beg += "\\hfill"

            # Establish font face and shape
            if font.get_type_face() == FONT_SANS_SERIF:
                thisstyle.font_beg += "\\sffamily"
                thisstyle.font_end = "\\rmfamily" + thisstyle.font_end
            if font.get_bold():
                thisstyle.font_beg += "\\bfseries"
                thisstyle.font_end = "\\mdseries" + thisstyle.font_end
            if font.get_italic() or font.get_underline():
                thisstyle.font_beg += "\\itshape"
                thisstyle.font_end = "\\upshape" + thisstyle.font_end

            # Now determine font size
            fontsize = map_font_size(size)
            if fontsize:
                thisstyle.font_beg += "\\" + fontsize
                thisstyle.font_end += "\\normalsize"

            thisstyle.font_beg += " "
            thisstyle.font_end += " "

            left = style.get_left_margin()
            first = style.get_first_indent() + left
            thisstyle.left_indent = left
            thisstyle.first_line_indent = first
            self.latexstyle[style_name] = thisstyle

    def close(self):
        """Clean up and close the document"""
        if self.in_list:
            self.emit("\\end{list}\n")
        self.emit("\\end{document}\n")
        self._backend.close()

    def end_page(self):
        """Issue a new page command"""
        self.emit("\\newpage")

    def start_paragraph(self, style_name, leader=None):
        """Paragraphs handling - A Gramps paragraph is any
        single body of text from a single word to several sentences.
        We assume a linebreak at the end of each paragraph."""
        style_sheet = self.get_style_sheet()

        style = style_sheet.get_paragraph_style(style_name)
        ltxstyle = self.latexstyle[style_name]
        self.level = style.get_header_level()

        self.fbeg = ltxstyle.font_beg
        self.fend = ltxstyle.font_end

        self.indent = ltxstyle.left_indent
        self.first_line_indent = ltxstyle.first_line_indent
        if self.indent == 0:
            self.indent = self.first_line_indent

        # For additional vertical space beneath title line(s)
        # i.e. when the first centering ended:
        if self.in_title and ltxstyle.font_beg.find("centering") == -1:
            self.in_title = False
            self._backend.write("\\vspace{5ex}%\n")
        if self.in_table:  #   paragraph in table indicates: cols of rows
            self.in_multrow_cell = True
        else:
            if leader:
                self._backend.write("".join(("\\grprepleader{", leader, "}%\n")))
            else:
                self._backend.write("\\grprepnoleader%\n")

            #           -------------------------------------------------------------------
            #   Gramps presumes 'cm' as units; here '\\grbaseindent' is used
            #   as equivalent, set in '_LATEX_TEMPLATE' above to '3em';
            #   there another value might be choosen.
            #           -------------------------------------------------------------------
            if self.indent is not None:
                self._backend.write(
                    "".join(
                        (
                            "\\grminpghead{",
                            repr(self.indent),
                            "}{",
                            repr(self.pict_width),
                            "}%\n",
                        )
                    )
                )
                self.fix_indent = True

                if leader is not None and not self.in_list:
                    self.in_list = True
                    self._backend.write("".join(("\\grlisthead{", leader, "}%\n")))

        if leader is None:
            self.emit("\n")
            self.emit("%s " % self.fbeg)

    def end_paragraph(self):
        """End the current paragraph"""
        newline = "%\n\n"
        if self.in_list:
            self.in_list = False
            self.emit("\n\\grlisttail%\n")
            newline = ""
        elif self.in_table:
            newline = SEPARATION_PAT

        self.emit("%s%s" % (self.fend, newline))
        if self.fix_indent:
            self.emit("\\grminpgtail%\n\n")
            self.fix_indent = False

        if self.pict_width:
            self.pict_width = 0
            self.pict_height = 0

    def start_bold(self):
        """Bold face"""
        self.emit("\\textbf{")

    def end_bold(self):
        """End bold face"""
        self.emit("}")

    def start_superscript(self):
        self.emit("\\textsuperscript{")

    def end_superscript(self):
        self.emit("}")

    def start_table(self, name, style_name):
        """Begin new table"""
        self.in_table = True
        self.currow = 0

        # We need to know a priori how many columns are in this table
        styles = self.get_style_sheet()
        self.tblstyle = styles.get_table_style(style_name)
        self.numcols = self.tblstyle.get_columns()
        self.column_order = []
        for cell in range(self.numcols):
            self.column_order.append(cell)
        if self.get_rtl_doc():
            self.column_order.reverse()

        tblfmt = "*{%d}{l}" % self.numcols
        self.emit("\\begin{longtable}[l]{%s}\n" % (tblfmt), TAB_BEG)

    def end_table(self):
        """Close the table environment"""
        self.emit("%\n\\end{longtable}%\n", TAB_END)
        self.in_table = False

    def start_row(self):
        """Begin a new row"""
        self.emit("", ROW_BEG)
        # doline/skipfirst are flags for adding hor. rules
        self.doline = False
        self.skipfirst = False
        self.curcol = 0
        self.currow += 1

    def end_row(self):
        """End the row (new line)"""
        self.emit("\\\\ ")
        if self.doline:
            if self.skipfirst:
                self.emit("".join((("\\cline{2-%d}" % self.numcols), "%\n")), ROW_END)
            else:
                self.emit("\\hline %\n", ROW_END)
        else:
            self.emit("%\n", ROW_END)
        self.emit("%\n")

    def start_cell(self, style_name, span=1):
        """Add an entry to the table.
        We always place our data inside braces
        for safety of formatting."""
        self.colspan = span
        self.curcol += self.colspan

        styles = self.get_style_sheet()
        self.cstyle = styles.get_cell_style(style_name)

        #       ------------------------------------------------------------------
        # begin special modification for boolean values
        # values imported here are used for test '==1' and '!=0'. To get
        # local boolean values the tests are now transfered to the import lines
        #       ------------------------------------------------------------------
        self.lborder = self.cstyle.get_left_border() == 1
        self.rborder = self.cstyle.get_right_border() == 1
        self.bborder = self.cstyle.get_bottom_border() == 1
        self.tborder = self.cstyle.get_top_border() != 0

        # self.llist not needed any longer.
        # now column widths are arranged in self.calc_latex_widths()
        # serving for fitting of cell contents at any column position.
        # self.llist = 1 == self.cstyle.get_longlist()

        cellfmt = "l"
        # Account for vertical rules
        if self.lborder:
            cellfmt = "|" + cellfmt
        if self.rborder:
            cellfmt = cellfmt + "|"

        # and Horizontal rules
        if self.bborder:
            self.doline = True
        elif self.curcol == 1:
            self.skipfirst = True
        if self.tborder:
            self.head_line = True
        #       ------------------------------------------------------------------
        #         end special modification for boolean values
        #       ------------------------------------------------------------------

        self.emit("\\multicolumn{%d}{%s}" % (span, cellfmt), CELL_BEG, span)

    def end_cell(self):
        """Prepares for next cell"""
        self.emit("", CELL_END)

    def add_media(self, infile, pos, x, y, alt="", style_name=None, crop=None):
        """Add photo to report"""
        outfile = os.path.splitext(infile)[0]
        pictname = latexescape(os.path.split(outfile)[1])
        outfile = "".join((outfile, ".jpg"))
        outfile2 = "".join((outfile, ".jpeg"))
        outfile3 = "".join((outfile, ".png"))
        if HAVE_PIL and infile not in [outfile, outfile2, outfile3]:
            try:
                curr_img = Image.open(infile)
                curr_img.save(outfile)
                width, height = curr_img.size
                if height > width:
                    y = y * height / width
            except IOError:
                self.emit(
                    "".join(
                        (
                            "%\n *** Error: cannot convert ",
                            infile,
                            "\n ***                    to ",
                            outfile,
                            "%\n",
                        )
                    )
                )
        elif not HAVE_PIL:
            from gramps.gen.config import config

            if not config.get("interface.ignore-pil"):
                from gramps.gen.constfunc import has_display

                if has_display() and self.uistate:
                    from gramps.gui.dialog import MessageHideDialog

                    title = _("PIL (Python Imaging Library) not loaded.")
                    message = _(
                        "Production of jpg images from non-jpg images "
                        "in LaTeX documents will not be available. "
                        "Use your package manager to install "
                        "python-imaging or python-pillow or "
                        "python3-pillow"
                    )
                    MessageHideDialog(
                        title,
                        message,
                        "interface.ignore-pil",
                        parent=self.uistate.window,
                    )
            self.emit(
                "".join(
                    (
                        "%\n *** Error: cannot convert ",
                        infile,
                        "\n ***                    to ",
                        outfile,
                        "\n *** PIL not installed %\n",
                    )
                )
            )

        if self.in_table:
            self.pict_in_table = True

        self.emit(
            "".join(
                (
                    "\\grmkpicture{",
                    outfile,
                    "}{",
                    repr(x),
                    "}{",
                    repr(y),
                    "}{",
                    pictname,
                    "}%\n",
                )
            )
        )
        self.pict_width = x
        self.pict_height = y

    def write_text(self, text, mark=None, links=False):
        """Write the text to the file"""
        if text == "\n":
            text = ""
        text = latexescape(text)

        if links is True:
            text = re.sub(URL_PATTERN, _CLICKABLE, text)

        # hard coded replace of the underline used for missing names/data
        text = text.replace("\\_" * 13, "\\underline{\\hspace{3\\grbaseindent}}")
        self.emit(text + " ")

    def write_styled_note(
        self, styledtext, format, style_name, contains_html=False, links=False
    ):
        """
        Convenience function to write a styledtext to the latex doc.
        styledtext : assumed a StyledText object to write
        format : = 0 : Flowed, = 1 : Preformatted
        style_name : name of the style to use for default presentation
        contains_html: bool, the backend should not check if html is present.
            If contains_html=True, then the textdoc is free to handle that in
            some way. Eg, a textdoc could remove all tags, or could make sure
            a link is clickable. self ignores notes that contain html
        links: bool, make URLs clickable if True
        """
        if contains_html:
            return
        text = str(styledtext)

        s_tags = styledtext.get_tags()
        if format:
            # preformatted, use different escape function
            self._backend.setescape(True)

        markuptext = self._backend.add_markup_from_styled(text, s_tags)

        if links is True:
            markuptext = re.sub(URL_PATTERN, _CLICKABLE, markuptext)
        markuptext = self._backend.add_markup_from_styled(text, s_tags)

        # there is a problem if we write out a note in a table.
        # ..................
        # now solved by postprocessing in self.calc_latex_widths()
        # by explicitely setting suitable width for all columns.
        #
        if format:
            self.start_paragraph(style_name)
            self.emit(markuptext)
            self.end_paragraph()
            # preformatted finished, go back to normal escape function
            self._backend.setescape(False)
        else:
            for line in markuptext.split("%\n%\n "):
                self.start_paragraph(style_name)
                for realline in line.split("\n"):
                    self.emit(realline)
                    self.emit("~\\newline \n")
                self.end_paragraph()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gramps.gen.plug._pluginreg import newplugin, STABLE
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

MODULE_VERSION = "5.2"

# ------------------------------------------------------------------------
#
# Ascii docgen
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "asciidoc"
plg.name = _("Plain Text")
plg.description = _("Generates documents in plain text format (.txt).")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "asciidoc.py"
plg.ptype = DOCGEN
plg.docclass = "AsciiDoc"
plg.optionclass = "AsciiDocOptions"
plg.paper = True
plg.style = True
plg.extension = "txt"

# ------------------------------------------------------------------------
#
# GTKPrint docgen
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "gtkprint"
plg.name = _("Print...")
plg.description = _("Generates documents and prints them directly.")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "gtkprint.py"
plg.ptype = DOCGEN
plg.docclass = "GtkPrint"
plg.optionclass = None
plg.paper = True
plg.style = True
plg.extension = ""

# ------------------------------------------------------------------------
#
# HtmlDoc docgen
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "htmldoc"
plg.name = _("HTML")
plg.description = _("Generates documents in HTML format.")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "htmldoc.py"
plg.ptype = DOCGEN
plg.docclass = "HtmlDoc"
plg.optionclass = None
plg.paper = False
plg.style = True
plg.extension = "html"

# ------------------------------------------------------------------------
#
# LaTeXDoc docgen
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "latexdoc"
plg.name = _("LaTeX")
plg.description = _("Generates documents in LaTeX format.")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "latexdoc.py"
plg.ptype = DOCGEN
plg.docclass = "LaTeXDoc"
plg.optionoclass = None
plg.paper = True
plg.style = False
plg.extension = "tex"

# ------------------------------------------------------------------------
#
# ODFDoc docgen
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "odfdoc"
plg.name = _("OpenDocument Text")
plg.description = _("Generates documents in OpenDocument " "Text format (.odt).")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "odfdoc.py"
plg.ptype = DOCGEN
plg.docclass = "ODFDoc"
plg.optionclass = None
plg.paper = True
plg.style = True
plg.extension = "odt"

# ------------------------------------------------------------------------
#
# PdfDoc docgen
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "pdfdoc"
plg.name = _("PDF document")
plg.description = _("Generates documents in PDF format (.pdf).")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "cairodoc.py"
plg.ptype = DOCGEN
plg.docclass = "PdfDoc"
plg.optionclass = None
plg.paper = True
plg.style = True
plg.extension = "pdf"

# ------------------------------------------------------------------------
#
# PSDrawDoc docgen
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "psdrawdoc"
plg.name = _("PostScript")
plg.description = _("Generates documents in PostScript format (.ps).")
plg.version = "2.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "cairodoc.py"
plg.ptype = DOCGEN
plg.docclass = "PsDoc"
plg.optionclass = None
plg.paper = True
plg.style = True
plg.extension = "ps"

# ------------------------------------------------------------------------
#
# RTFDoc docgen
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "rftdoc"
plg.name = _("RTF document")
plg.description = _("Generates documents in Rich Text format (.rtf).")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "rtfdoc.py"
plg.ptype = DOCGEN
plg.docclass = "RTFDoc"
plg.optionclass = None
plg.paper = True
plg.style = True
plg.extension = "rtf"

# ------------------------------------------------------------------------
#
# SvgDrawDoc docgen
#
# ------------------------------------------------------------------------

plg = newplugin()
plg.id = "SVG (Scalable Vector Graphics)"
plg.name = _("SVG document")
plg.description = _("Generates documents in Scalable " "Vector Graphics format (.svg).")
plg.version = "1.0"
plg.gramps_target_version = MODULE_VERSION
plg.status = STABLE
plg.fname = "svgdrawdoc.py"
plg.ptype = DOCGEN
plg.docclass = "SvgDrawDoc"
plg.optionclass = "SvgDrawDocOptions"
plg.paper = True
plg.style = True
plg.extension = "svg"



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2020 Paul Culley <paulr2787@gmail.com>
# Copyright (C) 2020  Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
""" BSDDB upgrade module """
# ------------------------------------------------------------------------
#
# Python Modules
#
# ------------------------------------------------------------------------
import os
import pickle
import logging
from bsddb3.db import DB, DB_DUP, DB_HASH, DB_RDONLY

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.plugins.db.dbapi.sqlite import SQLite
from gramps.cli.clidbman import NAME_FILE, find_next_db_dir
from gramps.gen.db.dbconst import DBBACKEND, DBMODE_W, SCHVERSFN
from gramps.gen.db.exceptions import (
    DbException,
    DbSupportedError,
    DbUpgradeRequiredError,
    DbVersionError,
)
from gramps.gen.db.utils import clear_lock_file
from gramps.gen.lib import Researcher
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.updatecallback import UpdateCallback

_ = glocale.translation.gettext

LOG = logging.getLogger(".upgrade")
_MINVERSION = 9
_DBVERSION = 19


class DbBsddb(SQLite):
    """
    Gramps BSDDB Converter
    """

    def __init__(self):
        """Create a new GrampsDB."""

        super().__init__()

    def load(
        self,
        dirname,
        callback=None,
        mode=DBMODE_W,
        force_schema_upgrade=False,
        update=True,
        username=None,
        password=None,
    ):
        """
        Here we create a sqlite db, and copy the bsddb into it.
        The new db is initially in a new directory, when we finish the copy
        we replace the contents of the original directory with the new db.

        We alway raise an exception to complete this, as the new db still
        needs to be upgraded some more.  When we raise the exception, the new
        db is closed.
        """
        if not update:
            raise DbException("Not Available")
        if not force_schema_upgrade:  # make sure user wants to upgrade
            raise DbSupportedError(_("BSDDB"))

        UpdateCallback.__init__(self, callback)

        # Here we open the dbapi db (a new one) for writing
        new_path = find_next_db_dir()
        os.mkdir(new_path)
        # store dbid in new dir
        dbid = "sqlite"
        backend_path = os.path.join(new_path, DBBACKEND)
        with open(backend_path, "w", encoding="utf8") as backend_file:
            backend_file.write(dbid)

        super().load(
            new_path,
            callback=None,
            mode="w",
            force_schema_upgrade=False,
            username=username,
            password=password,
        )

        # now read in the bsddb and copy to dpapi
        schema_vers = None
        total = 0
        tables = (
            ("person", "person"),
            ("family", "family"),
            ("event", "event"),
            ("place", "place"),
            ("repo", "repository"),
            ("source", "source"),
            ("citation", "citation"),
            ("media", "media"),
            ("note", "note"),
            ("tag", "tag"),
            ("meta_data", "metadata"),
        )

        # open each dbmap, and get its length for the total
        file_name = os.path.join(dirname, "name_group.db")
        if os.path.isfile(file_name):
            name_group_dbmap = DB()
            name_group_dbmap.set_flags(DB_DUP)
            name_group_dbmap.open(file_name, "name_group", DB_HASH, DB_RDONLY)
            total += len(name_group_dbmap)
        else:
            name_group_dbmap = None

        table_list = []
        for old_t, new_t in tables:
            file_name = os.path.join(dirname, old_t + ".db")
            if not os.path.isfile(file_name):
                continue
            dbmap = DB()
            dbmap.open(file_name, old_t, DB_HASH, DB_RDONLY)
            total += len(dbmap)
            table_list.append((old_t, new_t, dbmap))

        self.set_total(total)
        # copy data from each dbmap to sqlite table
        for old_t, new_t, dbmap in table_list:
            self._txn_begin()
            if new_t == "metadata":
                sql = "REPLACE INTO metadata (setting, value) VALUES " "(?, ?)"
            else:
                sql = "INSERT INTO %s (handle, blob_data) VALUES " "(?, ?)" % new_t

            for key in dbmap.keys():
                self.update()
                data = pickle.loads(dbmap[key], encoding="utf-8")

                if new_t == "metadata":
                    if key == b"version":
                        # found a schema version in metadata
                        schema_vers = data
                    elif key == b"researcher":
                        if len(data[0]) == 7:  # Pre-3.3 format
                            # Upgrade researcher data to include a locality
                            # field in the address.
                            addr = tuple([data[0][0], ""] + list(data[0][1:]))
                            new_data = (addr, data[1], data[2], data[3])
                        else:
                            new_data = data
                        data = Researcher().unserialize(new_data)
                    elif key == b"name_formats":
                        # upgrade formats if they were saved in the old way
                        for format_ix in range(len(data)):
                            fmat = data[format_ix]
                            if len(fmat) == 3:
                                fmat = fmat + (True,)
                                data[format_ix] = fmat
                    elif key == b"gender_stats":
                        # data is a dict, containing entries (see GenderStats)
                        self.dbapi.execute("DELETE FROM gender_stats")
                        g_sql = (
                            "INSERT INTO gender_stats "
                            "(given_name, female, male, unknown) "
                            "VALUES (?, ?, ?, ?)"
                        )
                        for name in data:
                            female, male, unknown = data[name]
                            self.dbapi.execute(g_sql, [name, female, male, unknown])
                        continue  # don't need this in metadata anymore
                    elif key == b"default":
                        # convert to string and change key
                        if isinstance(data, bytes):
                            data = data.decode("utf-8")
                        key = b"default-person-handle"
                    elif key == b"mediapath":
                        # change key
                        key = b"media-path"
                    elif key in [
                        b"surname_list",  # created by db now
                        b"pevent_names",  # obsolete
                        b"fevent_names",
                    ]:  # obsolete
                        continue
                    elif (
                        b"_names" in key
                        or b"refs" in key
                        or b"_roles" in key
                        or b"rels" in key
                        or b"_types" in key
                    ):
                        # These are list, but need to be set
                        data = set(data)

                self.dbapi.execute(sql, [key.decode("utf-8"), pickle.dumps(data)])

            # get schema version from file if not in metadata
            if new_t == "metadata" and schema_vers is None:
                versionpath = os.path.join(dirname, str(SCHVERSFN))
                if os.path.isfile(versionpath):
                    with open(versionpath, "r") as version_file:
                        schema_vers = int(version_file.read().strip())
                else:
                    schema_vers = 0
                # and put schema version into metadata
                self.dbapi.execute(sql, ["version", schema_vers])
            self._txn_commit()
            dbmap.close()
            if new_t == "metadata" and schema_vers < _MINVERSION:
                raise DbVersionError(schema_vers, _MINVERSION, _DBVERSION)

        if name_group_dbmap:
            self._txn_begin()
            for key in name_group_dbmap.keys():
                self.update()
                # name_group data (grouping) is NOT pickled
                data = name_group_dbmap[key]
                name = key.decode("utf-8")
                grouping = data.decode("utf-8")
                self.dbapi.execute(
                    "INSERT INTO name_group (name, grouping) VALUES (?, ?)",
                    [name, grouping],
                )
            self._txn_commit()
            name_group_dbmap.close()

        # done with new sqlite db, close it.  Cannot use normal close as it
        # overwrites the metadata.
        self._close()
        try:
            clear_lock_file(self.get_save_path())
        except IOError:
            pass
        self.db_is_open = False
        self._directory = None

        # copy tree name to new dir
        old_db_name = os.path.join(dirname, NAME_FILE)
        db_name = os.path.join(new_path, NAME_FILE)
        with open(old_db_name, "r", encoding="utf8") as _file:
            name = _file.read().strip()
        with open(db_name, "w", encoding="utf8") as _file:
            _file.write(name)
        # remove files from old dir
        for filename in os.listdir(dirname):
            file_path = os.path.join(dirname, filename)
            try:
                os.unlink(file_path)
            except Exception as e:
                LOG.error("Failed to delete %s. Reason: %s" % (file_path, e))
        # copy new db files to old dir
        for filename in os.listdir(new_path):
            old_file_path = os.path.join(new_path, filename)
            file_path = os.path.join(dirname, filename)
            try:
                os.replace(old_file_path, file_path)
            except Exception as e:
                LOG.error("Failed to move %s. Reason: %s" % (old_file_path, e))
        os.rmdir(new_path)

        # done preparing new db, but we still need to finish schema upgrades
        raise DbUpgradeRequiredError(schema_vers, "xx")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2006  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2011-2012  Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

""" The ReportDialog base class """

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import os

import logging

LOG = logging.getLogger(".")

# -------------------------------------------------------------------------
#
# GTK+ modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.config import config
from gramps.gen.const import URL_MANUAL_PAGE, DOCGEN_OPTIONS
from gramps.gen.errors import DatabaseError, FilterError, ReportError, WindowActiveError
from ...utils import open_file_with_default_application
from .. import add_gui_options, make_gui_option
from ...user import User
from ...dialog import ErrorDialog, OptionDialog
from gramps.gen.plug.report import (
    CATEGORY_TEXT,
    CATEGORY_DRAW,
    CATEGORY_BOOK,
    CATEGORY_CODE,
    CATEGORY_WEB,
    CATEGORY_GRAPHVIZ,
    CATEGORY_TREE,
    standalone_categories,
)
from gramps.gen.plug.docgen import StyleSheet, StyleSheetList
from ...managedwindow import ManagedWindow
from ._stylecombobox import StyleComboBox
from ._styleeditor import StyleListDisplay
from ._fileentry import FileEntry
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
# -------------------------------------------------------------------------
#
# Private Constants
#
# -------------------------------------------------------------------------
URL_REPORT_PAGE = URL_MANUAL_PAGE + "_-_Reports"


# -------------------------------------------------------------------------
#
# ReportDialog class
#
# -------------------------------------------------------------------------
class ReportDialog(ManagedWindow):
    """
    The ReportDialog base class.  This is a base class for generating
    customized dialogs to solicit options for a report.  It cannot be
    used as is, but it can be easily sub-classed to create a functional
    dialog for a stand-alone report.
    """

    border_pad = 6

    def __init__(self, dbstate, uistate, option_class, name, trans_name, track=[]):
        """Initialize a dialog to request that the user select options
        for a basic *stand-alone* report."""

        self.style_name = "default"
        self.firstpage_added = False
        self.raw_name = name
        self.dbstate = dbstate
        self.uistate = uistate
        self.db = dbstate.db
        self.report_name = trans_name

        ManagedWindow.__init__(self, uistate, track, self)

        self.init_options(option_class)
        self.init_interface()

    def close(self, *obj):
        """
        Close itself.
        cleanup things that can prevent garbage collection
        """
        if hasattr(self, "widgets"):  # handle pathlogical bug 4145
            totwidg = list(range(len(self.widgets)))
            totwidg.reverse()
            for ind in totwidg:
                if hasattr(self.widgets[ind][1], "clean_up"):
                    self.widgets[ind][1].clean_up()
                del self.widgets[ind]
            delattr(self, "widgets")
        for name, fram in self.frames.items():
            totwidg = list(range(len(fram)))
            totwidg.reverse()
            for ind in totwidg:
                if hasattr(fram[ind][1], "clean_up"):
                    fram[ind][1].clean_up()
                del fram[ind]
        self.frames.clear()
        self.frames = None
        ManagedWindow.close(self, *obj)

    def init_options(self, option_class):
        try:
            if issubclass(option_class, object):
                self.options = option_class(self.raw_name, self.db)
        except TypeError:
            self.options = option_class
        self.options.load_previous_values()

    def build_window_key(self, obj):
        key = self.raw_name
        return key

    def build_menu_names(self, obj):
        return (_("Configuration"), self.report_name)

    def init_interface(self):
        self.widgets = []
        self.doc_widgets = []
        self.frame_names = []
        self.frames = {}
        self.format_menu = None
        self.style_button = None

        self.style_name = self.options.handler.get_default_stylesheet_name()

        window = Gtk.Dialog(title="Gramps")
        self.set_window(window, None, self.get_title())
        self.window.set_modal(True)

        self.help = self.window.add_button(_("_Help"), Gtk.ResponseType.HELP)
        self.help.connect("clicked", self.on_help_clicked)

        self.cancel = self.window.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.cancel.connect("clicked", self.on_cancel)

        self.ok = self.window.add_button(_("_OK"), Gtk.ResponseType.OK)
        self.ok.connect("clicked", self.on_ok_clicked)

        self.window.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.window.set_default_size(600, -1)

        # Set up and run the dialog.  These calls are not in top down
        # order when looking at the dialog box as there is some
        # interaction between the various frames.

        self.setup_title()
        self.setup_header()
        self.grid = Gtk.Grid()
        self.grid.set_column_spacing(12)
        self.grid.set_row_spacing(6)
        self.grid.set_border_width(6)
        self.row = 0

        # Build the list of widgets that are used to extend the Options
        # frame and to create other frames
        self.add_user_options()

        self.setup_init()
        self.setup_format_frame()
        self.setup_target_frame()
        self.setup_style_frame()

        self.notebook = Gtk.Notebook()
        self.notebook.set_scrollable(True)
        self.notebook.set_border_width(6)
        try:
            # assume a vbox or hbox
            self.window.vbox.pack_start(
                self.notebook, expand=True, fill=True, padding=0
            )
        except:
            # general container instead:
            self.window.vbox.add(self.notebook)

        self.setup_report_options_frame()
        self.setup_other_frames()
        self.notebook.set_current_page(0)

        try:
            # assume a vbox or hbox
            self.window.vbox.pack_start(self.grid, expand=True, fill=True, padding=0)
        except:
            # general container instead:
            self.window.vbox.add(self.grid)
        self.show()

    def get_title(self):
        """The window title for this dialog"""
        name = self.report_name
        category = standalone_categories[self.category][1]
        return "%s - %s - Gramps" % (name, category)

    # ------------------------------------------------------------------------
    #
    # Functions related to extending the options
    #
    # ------------------------------------------------------------------------
    def add_user_options(self):
        """Called to allow subclasses to add widgets to the dialog form.
        It is called immediately before the window is displayed. All
        calls to add_option or add_frame_option should be called in
        this task."""
        add_gui_options(self)

    def parse_user_options(self):
        """Called to allow parsing of added widgets.
        It is called when OK is pressed in a dialog.
        All custom widgets should provide a parsing code here."""
        try:
            self.options.parse_user_options()
        except:
            LOG.error("Failed to parse user options.", exc_info=True)

    def add_option(self, label_text, widget):
        """Takes a text string and a Gtk Widget, and stores them to be
        appended to the Options section of the dialog. The text string
        is used to create a label for the passed widget. This allows the
        subclass to extend the Options section with its own widgets. The
        subclass is responsible for all managing of the widgets, including
        extracting the final value before the report executes. This task
        should only be called in the add_user_options task."""
        self.widgets.append((label_text, widget))

    def add_frame_option(self, frame_name, label_text, widget):
        """Similar to add_option this method takes a frame_name, a
        text string and a Gtk Widget. When the interface is built,
        all widgets with the same frame_name are grouped into a
        GtkFrame. This allows the subclass to create its own sections,
        filling them with its own widgets. The subclass is responsible for
        all managing of the widgets, including extracting the final value
        before the report executes. This task should only be called in
        the add_user_options task."""

        if frame_name in self.frames:
            self.frames[frame_name].append((label_text, widget))
        else:
            self.frames[frame_name] = [(label_text, widget)]
            self.frame_names.append(frame_name)

    # ------------------------------------------------------------------------
    #
    # Functions to create a default output style.
    #
    # ------------------------------------------------------------------------

    def build_style_menu(self, default=None):
        """Build a menu of style sets that are available for use in
        this report.  This menu will always have a default style
        available, and will have any other style set name that the
        user has previously created for this report.  This menu is
        created here instead of inline with the rest of the style
        frame, because it must be recreated to reflect any changes
        whenever the user closes the style editor dialog."""

        if default is None:
            default = self.style_name

        style_sheet_map = self.style_sheet_list.get_style_sheet_map()
        self.style_menu.set(style_sheet_map, default)

    # ------------------------------------------------------------------------
    #
    # Functions related to setting up the dialog window.
    #
    # ------------------------------------------------------------------------
    def setup_title(self):
        """Set up the title bar of the dialog.  This function relies
        on the get_title() customization function for what the title
        should be."""
        self.window.set_title(self.get_title())

    def setup_header(self):
        """Set up the header line bar of the dialog."""
        label = Gtk.Label(
            label='<span size="larger" weight="bold">%s</span>' % self.report_name
        )
        label.set_use_markup(True)
        self.window.vbox.pack_start(label, False, False, self.border_pad)

    def setup_style_frame(self):
        """Set up the style frame of the dialog.  This function relies
        on other routines to create the default style for this report,
        and to read in any user-defined styles for this report.  It
        then builds a menu of all the available styles for the user to
        choose from."""
        # Build the default style set for this report.
        self.default_style = StyleSheet()
        self.options.make_default_style(self.default_style)

        if self.default_style.is_empty():
            # Don't display the option if no styles are used
            return

        # Styles Frame
        label = Gtk.Label(label=_("%s:") % _("Style"))
        label.set_halign(Gtk.Align.START)

        self.style_menu = StyleComboBox()
        self.style_menu.set_hexpand(True)
        self.style_button = Gtk.Button(label="%s..." % _("Style Editor"))
        self.style_button.connect("clicked", self.on_style_edit_clicked)

        self.grid.attach(label, 1, self.row, 1, 1)
        self.grid.attach(self.style_menu, 2, self.row, 1, 1)
        self.grid.attach(self.style_button, 3, self.row, 1, 1)
        self.row += 1

        # Build the initial list of available styles sets.  This
        # includes the default style set and any style sets saved from
        # previous invocations of gramps.
        self.style_sheet_list = StyleSheetList(
            self.options.handler.get_stylesheet_savefile(), self.default_style
        )

        # Now build the actual menu.
        style = self.options.handler.get_default_stylesheet_name()
        self.build_style_menu(style)

    def setup_report_options_frame(self):
        """Set up the report options frame of the dialog.  This
        function relies on several report_xxx() customization
        functions to determine which of the items should be present in
        this box.  *All* of these items are optional, although the
        generations fields is used in most
        (but not all) dialog boxes."""

        row = 0
        max_rows = len(self.widgets)

        if max_rows == 0:
            return

        grid = Gtk.Grid()
        grid.set_border_width(6)
        grid.set_column_spacing(12)
        grid.set_row_spacing(6)

        label = Gtk.Label(label="<b>%s</b>" % _("Report Options"))
        label.set_halign(Gtk.Align.START)
        label.set_use_markup(True)

        self.notebook.append_page(grid, label)

        # Setup requested widgets
        for text, widget in self.widgets:
            widget.set_hexpand(True)
            if text:
                # Translators: needed for French, ignore otherwise
                text_widget = Gtk.Label(label=_("%s:") % text)
                text_widget.set_halign(Gtk.Align.START)
                grid.attach(text_widget, 1, row, 1, 1)
                grid.attach(widget, 2, row, 1, 1)
            else:
                grid.attach(widget, 2, row, 1, 1)
            row += 1

    def setup_other_frames(self):
        from .._guioptions import GuiTextOption

        for key in self.frame_names:
            flist = self.frames[key]
            grid = Gtk.Grid()
            grid.set_column_spacing(12)
            grid.set_row_spacing(6)
            grid.set_border_width(6)
            l = Gtk.Label(label="<b>%s</b>" % _(key))
            l.set_use_markup(True)
            self.notebook.append_page(grid, l)

            row = 0
            for text, widget in flist:
                widget.set_hexpand(True)
                if text:
                    text_widget = Gtk.Label(label=_("%s:") % text)
                    text_widget.set_halign(Gtk.Align.START)
                    grid.attach(text_widget, 1, row, 1, 1)
                    if isinstance(widget, GuiTextOption):
                        grid.attach(widget, 2, row, 1, 1)
                    else:
                        grid.attach(widget, 2, row, 1, 1)
                else:
                    grid.attach(widget, 2, row, 1, 1)
                row += 1

    # ------------------------------------------------------------------------
    #
    # Customization hooks for stand-alone reports (subclass ReportDialog)
    #
    # ------------------------------------------------------------------------
    def setup_format_frame(self):
        """Not used in bare report dialogs. Override in the subclass."""
        pass

    # ------------------------------------------------------------------------
    #
    # Functions related getting/setting the default directory for a dialog.
    #
    # ------------------------------------------------------------------------
    def get_default_directory(self):
        """Get the name of the directory to which the target dialog
        box should default.  This value can be set in the preferences
        panel."""
        return config.get("paths.report-directory")

    def set_default_directory(self, value):
        """Save the name of the current directory, so that any future
        reports will default to the most recently used directory.
        This also changes the directory name that will appear in the
        preferences panel, but does not change the preference in disk.
        This means that the last directory used will only be
        remembered for this session of gramps unless the user saves
        his/her preferences."""
        config.set("paths.report-directory", value)

    # ------------------------------------------------------------------------
    #
    # Functions related to setting up the dialog window.
    #
    # ------------------------------------------------------------------------
    def setup_init(self):
        # add any elements that we are going to need:
        hid = self.style_name
        if hid[-4:] == ".xml":
            hid = hid[0:-4]
        self.target_fileentry = FileEntry(hid, _("Save As"), parent=self.window)
        spath = self.get_default_directory()
        self.target_fileentry.set_filename(spath)
        # need any labels at top:
        label = Gtk.Label(label="<b>%s</b>" % _("Document Options"))
        label.set_use_markup(1)
        label.set_halign(Gtk.Align.START)
        self.grid.set_border_width(12)
        self.grid.attach(label, 0, self.row, 4, 1)
        self.row += 1

    def setup_target_frame(self):
        """Set up the target frame of the dialog.  This function
        relies on several target_xxx() customization functions to
        determine whether the target is a directory or file, what the
        title of any browser window should be, and what default
        directory should be used."""

        # Save Frame
        self.doc_label = Gtk.Label(label=_("%s:") % _("Filename"))
        self.doc_label.set_halign(Gtk.Align.START)

        self.grid.attach(self.doc_label, 1, self.row, 1, 1)
        self.target_fileentry.set_hexpand(True)
        self.grid.attach(self.target_fileentry, 2, self.row, 2, 1)
        self.row += 1

    # ------------------------------------------------------------------------
    #
    # Functions related to retrieving data from the dialog window
    #
    # ------------------------------------------------------------------------
    def parse_target_frame(self):
        """Parse the target frame of the dialog.  If the target
        filename is empty this routine returns a special value of None
        to tell the calling routine to give up.  This function also
        saves the current directory so that any future reports will
        default to the most recently used directory."""
        self.target_path = self.target_fileentry.get_full_path(0)
        if not self.target_path:
            return None

        # First we check whether the selected path exists
        if os.path.exists(self.target_path):
            # selected path is an existing dir and we need a dir
            if os.path.isdir(self.target_path):
                # check whether the dir has rwx permissions
                if not os.access(self.target_path, os.R_OK | os.W_OK | os.X_OK):
                    ErrorDialog(
                        _("Permission problem"),
                        _(
                            "You do not have permission to write "
                            "under the directory %s\n\n"
                            "Please select another directory or correct "
                            "the permissions."
                        )
                        % self.target_path,
                        parent=self.window,
                    )
                    return None

            # selected path is an existing file and we need a file
            if os.path.isfile(self.target_path):
                aaa = OptionDialog(
                    _("File already exists"),
                    _(
                        "You can choose to either overwrite the "
                        "file, or change the selected filename."
                    ),
                    _("_Overwrite"),
                    None,
                    _("_Change filename"),
                    None,
                    parent=self.window,
                )

                if aaa.get_response() == Gtk.ResponseType.YES:
                    return None

        # selected path does not exist yet
        else:
            # we will need to create the file/dir
            # need to make sure we can create in the parent dir
            parent_dir = os.path.dirname(os.path.normpath(self.target_path))
            if os.path.isdir(parent_dir):
                if not os.access(parent_dir, os.W_OK):
                    ErrorDialog(
                        _("Permission problem"),
                        _(
                            "You do not have permission to create "
                            "%s\n\n"
                            "Please select another path or correct "
                            "the permissions."
                        )
                        % self.target_path,
                        parent=self.window,
                    )
                    return None
            else:
                ErrorDialog(
                    _("No directory"),
                    _(
                        "There is no directory %s.\n\n"
                        "Please select another directory "
                        "or create it."
                    )
                    % parent_dir,
                    parent=self.window,
                )
                return None

        self.set_default_directory(os.path.dirname(self.target_path) + os.sep)
        self.options.handler.output = self.target_path
        return 1

    def parse_style_frame(self):
        """Parse the style frame of the dialog.  Save the user
        selected output style for later use.  Note that this routine
        retrieves a value whether or not the menu is displayed on the
        screen.  The subclass will know whether this menu was enabled.
        This is for simplicity of programming."""
        if not self.default_style.is_empty():
            (style_name, self.selected_style) = self.style_menu.get_value()
            self.options.handler.set_default_stylesheet_name(style_name)

    # ------------------------------------------------------------------------
    #
    # Callback functions from the dialog
    #
    # ------------------------------------------------------------------------
    def on_ok_clicked(self, obj):
        """The user is satisfied with the dialog choices.  Validate
        the output file name before doing anything else.  If there is
        a file name, gather the options and create the report."""

        # Is there a filename?  This should also test file permissions, etc.
        if not self.parse_target_frame():
            self.window.run()

        # Preparation
        self.parse_style_frame()
        self.parse_user_options()

        # Save options
        self.options.handler.save_options()

    def on_cancel(self, *obj):
        pass

    def on_help_clicked(self, *obj):
        from ...display import display_help, display_url

        if hasattr(self.options, "help_url"):
            display_url(self.options.help_url)
            return
        display_help(URL_REPORT_PAGE, self.report_name.replace(" ", "_"))

    def on_style_edit_clicked(self, *obj):
        """The user has clicked on the 'Edit Styles' button.  Create a
        style sheet editor object and let them play.  When they are
        done, the previous routine will be called to update the dialog
        menu for selecting a style."""
        StyleListDisplay(
            self.style_sheet_list,
            self.uistate,
            self.track,
            callback=self.build_style_menu,
        )

    # ----------------------------------------------------------------------
    #
    # Functions related to any docgen options for a dialog.
    #
    # ----------------------------------------------------------------------
    def setup_doc_options_frame(self):
        if self.doc_widgets:
            for option_widget in self.doc_widgets:
                self.grid.remove(option_widget)
            self.doc_widgets = []
        self.doc_options = None

        if not self.doc_option_class:
            return  # this docgen type has no options

        self.init_doc_options(self.doc_option_class)
        menu = self.doc_options.menu
        for name in menu.get_option_names(DOCGEN_OPTIONS):
            option = menu.get_option(DOCGEN_OPTIONS, name)
            # override option default with xml-saved value:
            if name in self.doc_options.options_dict:
                option.set_value(self.doc_options.options_dict[name])
            widget, has_label = make_gui_option(
                option, self.dbstate, self.uistate, self.track
            )
            if has_label:
                widget_text = Gtk.Label(label=(_("%s:") % option.get_label()))
                widget_text.set_halign(Gtk.Align.START)
                self.grid.attach(widget_text, 1, self.row, 1, 1)
                self.doc_widgets.append(widget_text)
            self.grid.attach(widget, 2, self.row, 2, 1)
            self.doc_widgets.append(widget)
            self.row += 1

    def init_doc_options(self, option_class):
        try:
            if issubclass(option_class, object):
                self.doc_options = option_class(self.raw_name, self.db)
        except TypeError:
            self.doc_options = option_class
        self.doc_options.load_previous_values()

    def parse_doc_options(self):
        """
        Called to allow parsing of added docgen widgets.
        It is called when OK is pressed in a dialog.
        """
        if not self.doc_options:
            return
        try:
            self.doc_options.parse_user_options()
            for opt in self.doc_options.options_dict:
                self.options.options_dict[opt] = [
                    self.basedocname,
                    self.doc_options.options_dict[opt],
                ]
        except:
            logging.warning("Failed to parse doc options")


# ------------------------------------------------------------------------
#
# Generic task function a standalone GUI report
#
# ------------------------------------------------------------------------
def report(
    dbstate,
    uistate,
    person,
    report_class,
    options_class,
    trans_name,
    name,
    category,
    require_active,
):
    """
    report - task starts the report. The plugin system requires that the
    task be in the format of task that takes a database and a person as
    its arguments.
    """
    if require_active and not person:
        ErrorDialog(
            _("Active person has not been set"),
            _("You must select an active person for this report to work " "properly."),
            parent=uistate.window,
        )
        return

    if category == CATEGORY_TEXT:
        from ._textreportdialog import TextReportDialog

        dialog_class = TextReportDialog
    elif category == CATEGORY_DRAW:
        from ._drawreportdialog import DrawReportDialog

        dialog_class = DrawReportDialog
    elif category == CATEGORY_GRAPHVIZ:
        from ._graphvizreportdialog import GraphvizReportDialog

        dialog_class = GraphvizReportDialog
    elif category == CATEGORY_TREE:
        from ._treereportdialog import TreeReportDialog

        dialog_class = TreeReportDialog
    elif category == CATEGORY_WEB:
        from ._webreportdialog import WebReportDialog

        dialog_class = WebReportDialog
    elif category in (CATEGORY_BOOK, CATEGORY_CODE):
        try:
            report_class(dbstate, uistate)
        except WindowActiveError:
            pass
        return
    else:
        dialog_class = ReportDialog

    dialog = dialog_class(dbstate, uistate, options_class, name, trans_name)

    while True:
        response = dialog.window.run()
        if response == Gtk.ResponseType.OK:
            dialog.close()
            try:
                user = User(uistate=uistate)
                my_report = report_class(dialog.db, dialog.options, user)
                my_report.doc.init()
                my_report.begin_report()
                my_report.write_report()
                my_report.end_report()

                # Web reports do not have a target frame
                # The GtkPrint generator can not be "opened"
                if (
                    hasattr(dialog, "open_with_app")
                    and dialog.open_with_app.get_property("sensitive") == True
                    and dialog.open_with_app.get_active()
                ):
                    out_file = dialog.options.get_output()
                    open_file_with_default_application(out_file, uistate)

            except FilterError as msg:
                (msg1, msg2) = msg.messages()
                ErrorDialog(msg1, msg2, parent=uistate.window)
            except IOError as msg:
                ErrorDialog(
                    _("Report could not be created"), str(msg), parent=uistate.window
                )
            except ReportError as msg:
                (msg1, msg2) = msg.messages()
                ErrorDialog(msg1, msg2, parent=uistate.window)
            except DatabaseError as msg:
                ErrorDialog(
                    _("Report could not be created"), str(msg), parent=uistate.window
                )
                #           The following except statement will catch all "NoneType" exceptions.
                #           This is useful for released code where the exception is most likely
                #           a corrupt database. But it is less useful for developing new reports
                #           where the exception is most likely a report bug.
                #            except AttributeError,msg:
                #                if str(msg).startswith("'NoneType' object has no attribute"):
                #                    # "'NoneType' object has no attribute ..." usually means
                #                    # database corruption
                #                    RunDatabaseRepair(str(msg),
                #                                      parent=self.window)
                #                else:
                #                    raise
                raise
            except:
                LOG.error("Failed to run report.", exc_info=True)
            break
        elif response == Gtk.ResponseType.CANCEL:
            dialog.close()
            break
        elif response == Gtk.ResponseType.DELETE_EVENT:
            # just stop, in ManagedWindow, delete-event is already coupled to
            # correct action.
            break

    # do needed cleanup
    dialog.db = None
    dialog.options = None
    if hasattr(dialog, "window"):
        delattr(dialog, "window")
    if hasattr(dialog, "notebook"):
        delattr(dialog, "notebook")
    del dialog



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011 Nick Hall
# Copyright (C) 2011 Tim G L Lyons
# Copyright (C) 2020 Matthias Kemmer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# ------------------------------------------------------------------------
#
# Gtk
#
# ------------------------------------------------------------------------
from gi.repository import Gtk

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug import Gramplet
from gramps.gui.widgets.styledtexteditor import StyledTextEditor
from gramps.gui.widgets import SimpleButton
from gramps.gen.lib import StyledText
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


class Notes(Gramplet):
    """
    Displays the notes for an object.
    """

    def init(self):
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)
        self.gui.WIDGET.show()

    def build_gui(self):
        """
        Build the GUI interface.
        """
        top = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        hbox = Gtk.Box()
        self.left = SimpleButton("go-previous", self.left_clicked)
        self.left.set_sensitive(False)
        hbox.pack_start(self.left, False, False, 0)
        self.right = SimpleButton("go-next", self.right_clicked)
        self.right.set_sensitive(False)
        hbox.pack_start(self.right, False, False, 0)
        self.page = Gtk.Label()
        self.page.set_halign(Gtk.Align.START)
        hbox.pack_start(self.page, True, True, 10)
        self.ntype = Gtk.Label()
        hbox.pack_start(self.ntype, False, False, 10)

        scrolledwindow = Gtk.ScrolledWindow()
        scrolledwindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.texteditor = StyledTextEditor()
        self.texteditor.set_editable(False)
        self.texteditor.set_wrap_mode(Gtk.WrapMode.WORD)
        scrolledwindow.add(self.texteditor)

        top.pack_start(hbox, False, False, 0)
        top.pack_start(scrolledwindow, True, True, 0)
        top.show_all()
        return top

    def get_notes(self, obj):
        """
        Get the note list for the current object.
        """
        self.left.set_sensitive(False)
        self.right.set_sensitive(False)
        self.texteditor.set_text(StyledText())
        self.note_list = obj.get_note_list()
        self.page.set_text("")
        self.ntype.set_text("")
        if len(self.note_list) > 0:
            self.set_has_data(True)
            if len(self.note_list) > 1:
                self.right.set_sensitive(True)
            self.current = 0
            self.display_note()
        else:
            self.set_has_data(False)

    def clear_text(self):
        self.left.set_sensitive(False)
        self.right.set_sensitive(False)
        self.texteditor.set_text(StyledText())
        self.page.set_text("")
        self.current = 0

    def display_note(self):
        """
        Display the current note.
        """
        note_handle = self.note_list[self.current]
        note = self.dbstate.db.get_note_from_handle(note_handle)
        self.texteditor.set_text(note.get_styledtext())
        self.ntype.set_text(str(note.get_type()))
        self.page.set_text(
            _("%(current)d of %(total)d")
            % {"current": self.current + 1, "total": len(self.note_list)}
        )

    def left_clicked(self, button):
        """
        Display the previous note.
        """
        if self.current > 0:
            self.current -= 1
            self.right.set_sensitive(True)
            if self.current == 0:
                self.left.set_sensitive(False)
            self.display_note()

    def right_clicked(self, button):
        """
        Display the next note.
        """
        if self.current < len(self.note_list) - 1:
            self.current += 1
            self.left.set_sensitive(True)
            if self.current == len(self.note_list) - 1:
                self.right.set_sensitive(False)
            self.display_note()

    def get_has_data(self, obj):
        """
        Return True if the gramplet has data, else return False.
        """
        if obj is None:
            return False
        if obj.get_note_list():
            return True
        return False


class PersonNotes(Notes):
    """
    Displays the notes for a person.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "person-update", self.update)

    def active_changed(self, handle):
        self.update()

    def update_has_data(self):
        active_handle = self.get_active("Person")
        if active_handle:
            active = self.dbstate.db.get_person_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Person")
        if active_handle:
            active = self.dbstate.db.get_person_from_handle(active_handle)
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class EventNotes(Notes):
    """
    Displays the notes for an event.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "event-update", self.update)
        self.connect_signal("Event", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Event")
        if active_handle:
            active = self.dbstate.db.get_event_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Event")
        if active_handle:
            active = self.dbstate.db.get_event_from_handle(active_handle)
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class FamilyNotes(Notes):
    """
    Displays the notes for a family.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "family-update", self.update)
        self.connect_signal("Family", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Family")
        if active_handle:
            active = self.dbstate.db.get_family_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Family")
        if active_handle:
            active = self.dbstate.db.get_family_from_handle(active_handle)
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class PlaceNotes(Notes):
    """
    Displays the notes for a place.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "place-update", self.update)
        self.connect_signal("Place", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Place")
        if active_handle:
            active = self.dbstate.db.get_place_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Place")
        if active_handle:
            active = self.dbstate.db.get_place_from_handle(active_handle)
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class SourceNotes(Notes):
    """
    Displays the notes for a source.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "source-update", self.update)
        self.connect_signal("Source", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Source")
        if active_handle:
            active = self.dbstate.db.get_source_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Source")
        if active_handle:
            active = self.dbstate.db.get_source_from_handle(active_handle)
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class CitationNotes(Notes):
    """
    Displays the notes for a Citation.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "citation-update", self.update)
        self.connect_signal("Citation", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Citation")
        if active_handle:
            active = self.dbstate.db.get_citation_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Citation")
        if active_handle:
            active = self.dbstate.db.get_citation_from_handle(active_handle)
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class RepositoryNotes(Notes):
    """
    Displays the notes for a repository.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "repository-update", self.update)
        self.connect_signal("Repository", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Repository")
        if active_handle:
            active = self.dbstate.db.get_repository_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Repository")
        if active_handle:
            active = self.dbstate.db.get_repository_from_handle(active_handle)
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class MediaNotes(Notes):
    """
    Displays the notes for a media object.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "media-update", self.update)
        self.connect_signal("Media", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Media")
        if active_handle:
            active = self.dbstate.db.get_media_from_handle(active_handle)
            self.set_has_data(self.get_has_data(active))
        else:
            self.set_has_data(False)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Media")
        if active_handle:
            active = self.dbstate.db.get_media_from_handle(active_handle)
            if active:
                self.get_notes(active)
            else:
                self.set_has_data(False)
        else:
            self.set_has_data(False)


class NoteNotes(Notes):
    """
    Display a single note in NoteView.
    """

    def db_changed(self):
        self.connect(self.dbstate.db, "note-update", self.update)
        self.connect_signal("Note", self.update)

    def main(self):
        self.clear_text()
        active_handle = self.get_active("Note")
        if active_handle:
            active = self.dbstate.db.get_note_from_handle(active_handle)
            if active:
                self.texteditor.set_text(active.get_styledtext())



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009       B. Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# python
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

WEIGHT_NORMAL = Pango.Weight.NORMAL
WEIGHT_BOLD = Pango.Weight.BOLD

from html import escape

# -------------------------------------------------------------------------
#
# Gramps classes
#
# -------------------------------------------------------------------------
from ...widgets.undoablebuffer import UndoableBuffer
from gramps.gen.lib import EventRoleType, EventType, Date
from gramps.gen.datehandler import get_date, get_date_valid
from gramps.gen.config import config
from gramps.gen.utils.db import get_participant_from_event
from gramps.gen.display.place import displayer as place_displayer
from gramps.gen.proxy.cache import CacheProxyDb

# -------------------------------------------------------------------------
#
# Globals
#
# -------------------------------------------------------------------------
invalid_date_format = config.get("preferences.invalid-date-format")
age_precision = config.get("preferences.age-display-precision")
age_after_death = config.get("preferences.age-after-death")


# -------------------------------------------------------------------------
#
# EventRefModel
#
# -------------------------------------------------------------------------
class EventRefModel(Gtk.TreeStore):
    # index of the working group
    _ROOTINDEX = 0
    _GROUPSTRING = _("%(groupname)s - %(groupnumber)d")

    COL_DESCR = (0, str)
    COL_TYPE = (1, str)
    COL_GID = (2, str)
    COL_DATE = (3, str)
    COL_PLACE = (4, str)
    COL_ROLE = (5, str)
    COL_PARTIC = (6, str)
    COL_SORTDATE = (7, str)
    COL_EVENTREF = (8, object)
    COL_FONTWEIGHT = (9, int)
    COL_AGE = (10, str)
    COL_SORTAGE = (11, str)
    COL_PRIVATE = (12, bool)
    COL_HAS_SOURCE = (13, bool)

    COLS = (
        COL_DESCR,
        COL_TYPE,
        COL_GID,
        COL_DATE,
        COL_PLACE,
        COL_ROLE,
        COL_PARTIC,
        COL_SORTDATE,
        COL_EVENTREF,
        COL_FONTWEIGHT,
        COL_AGE,
        COL_SORTAGE,
        COL_PRIVATE,
        COL_HAS_SOURCE,
    )

    def __init__(self, event_list, db, groups, **kwargs):
        """
        @param event_list: A list of lists, every entry is a group, the entries
            in a group are the data that needs to be shown subordinate to the
            group
        @param db: a database objects that can be used to obtain info
        @param groups: a list of (key, name) tuples. key is a key for the group
            that might be used. name is the name for the group.
        @param kwargs: A dictionary of additional settings/values.
        """
        self.start_date = kwargs.get("start_date", None)
        self.end_date = kwargs.get("end_date", None)
        typeobjs = (x[1] for x in self.COLS)
        Gtk.TreeStore.__init__(self, *typeobjs)
        self.db = CacheProxyDb(db)
        self.groups = groups
        for index, group in enumerate(event_list):
            parentiter = self.append(None, row=self.row_group(index, group))
            for eventref in group:
                event = db.get_event_from_handle(eventref.ref)
                self.append(parentiter, row=self.row(index, eventref, event))

    def row_group(self, index, group):
        name = self.namegroup(index, len(group))
        spouse = self.groups[index][2]
        return [
            spouse,
            name,
            "",
            "",
            "",
            "",
            "",
            "",
            (index, None),
            WEIGHT_BOLD,
            "",
            "",
            None,
            None,
        ]

    def namegroup(self, groupindex, length):
        return self._GROUPSTRING % {
            "groupname": self.groups[groupindex][1],
            "groupnumber": length,
        }

    def row(self, index, eventref, event):
        return [
            event.get_description(),
            str(event.get_type()),
            event.get_gramps_id(),
            self.column_date(eventref),
            self.column_place(eventref),
            self.column_role(eventref),
            self.column_participant(eventref),
            self.column_sort_date(eventref),
            (index, eventref),
            self.colweight(index),
            self.column_age(event),
            self.column_sort_age(event),
            eventref.get_privacy(),
            event.has_citations(),
        ]

    def colweight(self, index):
        return WEIGHT_NORMAL

    def column_role(self, event_ref):
        return str(event_ref.get_role())

    def column_date(self, event_ref):
        event = self.db.get_event_from_handle(event_ref.ref)
        retval = get_date(event)
        if not get_date_valid(event):
            return invalid_date_format % escape(retval)
        else:
            return retval

    def column_sort_date(self, event_ref):
        event = self.db.get_event_from_handle(event_ref.ref)
        date = event.get_date_object()
        if date:
            return "%09d" % date.get_sort_value()
        else:
            return ""

    def column_place(self, event_ref):
        if event_ref and event_ref.ref:
            event = self.db.get_event_from_handle(event_ref.ref)
            if event:
                return place_displayer.display_event(self.db, event)
        return ""

    def column_participant(self, event_ref):
        return get_participant_from_event(self.db, event_ref.ref)

    def column_age(self, event):
        """
        Returns a string representation of age in years.  Change
        precision=2 for "year, month", or precision=3 for "year,
        month, days"
        """
        date = event.get_date_object()
        if date and self.start_date:
            if (
                date == self.start_date
                and date.modifier == Date.MOD_NONE
                and not (
                    event.get_type().is_death_fallback()
                    or event.get_type() == EventType.DEATH
                )
            ):
                return ""
            elif self.end_date and self.end_date < date and not age_after_death:
                return ""
            else:
                return (date - self.start_date).format(precision=age_precision)
        else:
            return ""

    def column_sort_age(self, event):
        """
        Returns a string version of number of days of age.
        """
        date = event.get_date_object()
        if date and self.start_date:
            return "%09d" % int(date - self.start_date)
        else:
            return ""



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2015 Douglas Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gramps.gen.plug._pluginreg import register, STABLE, DATABASE
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

register(
    DATABASE,
    id="sqlite",
    name=_("SQLite"),
    name_accell=_("_SQLite Database"),
    description=_("SQLite Database"),
    version="1.0.0",
    gramps_target_version="5.2",
    status=STABLE,
    fname="sqlite.py",
    databaseclass="SQLite",
    authors=["Doug Blank"],
    authors_email=["doug.blank@gmail.com"],
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008  Brian G. Matherly
# Copyright (C) 2008  Jerome Rapinat
# Copyright (C) 2008  Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# HasAddressText
#
# -------------------------------------------------------------------------
class HasAddressText(Rule):
    """Rule that checks for text in personal addresses"""

    labels = [_("Text:")]
    name = _("People with an address containing <text>")
    description = _(
        "Matches people with a personal address containing " "the given text"
    )
    category = _("General filters")
    allow_regex = True

    def apply(self, db, person):
        for address in person.get_address_list():
            for string in address.get_text_data_list():
                if self.match_substring(0, string):
                    return True
        return False



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Attempt to parse dates for Icelandic, Sveinn í Felli 2016

"""
Icelandic-specific classes for parsing and displaying dates.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re
import datetime

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler


# -------------------------------------------------------------------------
#
# Icelandic parser class
#
# -------------------------------------------------------------------------
class DateParserIs(DateParser):
    """
    Convert a text string into a Date object, expecting a date
    notation in the Icelandic language. If the date cannot be converted,
    the text string is assigned.
    """

    # modifiers before the date
    modifier_to_int = {
        "fyrir": Date.MOD_BEFORE,
        "á undan": Date.MOD_BEFORE,
        "eftir": Date.MOD_AFTER,
        "í kringum": Date.MOD_ABOUT,
        "uþb": Date.MOD_ABOUT,
        "um": Date.MOD_ABOUT,
        "from": Date.MOD_FROM,
        "to": Date.MOD_TO,
    }

    bce = ["f Kr"]

    calendar_to_int = {
        "gregoríanskt   ": Date.CAL_GREGORIAN,
        "g": Date.CAL_GREGORIAN,
        "júlíanskt": Date.CAL_JULIAN,
        "j": Date.CAL_JULIAN,
        "hebreskt": Date.CAL_HEBREW,
        "h": Date.CAL_HEBREW,
        "íslamskt": Date.CAL_ISLAMIC,
        "múslimskt": Date.CAL_ISLAMIC,
        "i": Date.CAL_ISLAMIC,
        "franskt": Date.CAL_FRENCH,
        "franska lýðveldisins": Date.CAL_FRENCH,
        "f": Date.CAL_FRENCH,
        "persneskt": Date.CAL_PERSIAN,
        "p": Date.CAL_PERSIAN,
        "sænskt": Date.CAL_SWEDISH,
        "s": Date.CAL_SWEDISH,
    }

    quality_to_int = {
        "áætlað": Date.QUAL_ESTIMATED,
        "reiknað": Date.QUAL_CALCULATED,
    }

    def dhformat_changed(self):
        self._dhformat_parse = re.compile(r".*%(\S).*%(\S).*%(\S).*%(\S).*")

    def init_strings(self):
        DateParser.init_strings(self)

        # match 'day. month year' format
        self._text2 = re.compile(
            r"(\d+)?\.?\s*?%s\.?\s*((\d+)(/\d+)?)?\s*$" % self._mon_str, re.IGNORECASE
        )
        # match 'short-day day.month year' format
        short_day_str = "(" + "|".join(self._ds.short_days[1:]) + ")"
        self._numeric = re.compile(
            r"%s\s*((\d+)[\.]\s*)?((\d+)\s*)?(\d+)\s*$" % short_day_str, re.IGNORECASE
        )
        self._span = re.compile(
            r"(frá)?\s*(?P<start>.+)\s*(til|--|–)\s*(?P<stop>.+)", re.IGNORECASE
        )
        self._range = re.compile(
            r"(milli)\s+(?P<start>.+)\s+og\s+(?P<stop>.+)", re.IGNORECASE
        )


# -------------------------------------------------------------------------
#
# Icelandic display class
#
# -------------------------------------------------------------------------
class DateDisplayIs(DateDisplay):
    """
    Icelandic language date display class.
    """

    long_months = (
        "",
        "janúar",
        "febrúar",
        "mars",
        "apríl",
        "maí",
        "júní",
        "júlí",
        "ágúst",
        "september",
        "október",
        "nóvember",
        "desember",
    )

    short_months = (
        "",
        "jan",
        "feb",
        "mar",
        "apr",
        "maí",
        "jún",
        "júl",
        "ágú",
        "sep",
        "okt",
        "nóv",
        "des",
    )

    formats = (
        "ÁÁÁÁ-MM-DD (ISO)",
        "Tölulegt",
        "Mánuður dagur, ár",
        "Mán Dag Ár",
        "Dagur mánuður ár",
        "Dag Mán Ár",
    )
    # this must agree with DateDisplayEn's "formats" definition
    # (since no locale-specific _display_gregorian exists, here)

    calendar = (
        "",
        "júlíanskt",
        "hebreskt",
        "franska lýðveldisins",
        "persneskt",
        "íslamskt",
        "sænskt",
    )

    _mod_str = ("", "fyrir ", "eftir ", "uþb ", "", "", "")

    _qual_str = ("", "reiknað ", "reiknað ")

    _bce_str = "%s f. Kr"

    def display(self, date):
        """
        Return a text string representing the date.
        """
        mod = date.get_modifier()
        cal = date.get_calendar()
        qual = date.get_quality()
        start = date.get_start_date()
        newyear = date.get_new_year()

        qual_str = self._qual_str[qual]

        if mod == Date.MOD_TEXTONLY:
            return date.get_text()
        elif start == Date.EMPTY:
            return ""
        elif mod == Date.MOD_SPAN:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%sfrá %s til %s%s" % (qual_str, d1, d2, scal)
        elif mod == Date.MOD_RANGE:
            d1 = self.display_cal[cal](start)
            d2 = self.display_cal[cal](date.get_stop_date())
            scal = self.format_extras(cal, newyear)
            return "%smilli %s og %s%s" % (qual_str, d1, d2, scal)
        else:
            text = self.display_cal[date.get_calendar()](start)
            scal = self.format_extras(cal, newyear)
            return "%s%s%s%s" % (qual_str, self._mod_str[mod], text, scal)

    def _get_weekday(self, date_val):
        if (
            date_val[0] == 0
            or date_val[1] == 0  # no day or no month or both
            or date_val[2] > datetime.MAXYEAR
        ):  # bug 10815
            return ""
        w_day = datetime.date(date_val[2], date_val[1], date_val[0])  # y, m, d
        return self.short_days[((w_day.weekday() + 1) % 7) + 1]

    def dd_dformat01(self, date_val):
        """
        numerical -- for Icelandic dates
        """
        if date_val[3]:
            return self.display_iso(date_val)
        else:
            if date_val[0] == date_val[1] == 0:
                return str(date_val[2])
            else:
                value = self.dhformat.replace("%m", str(date_val[1]))
                # some locales have %b for the month, e.g. ar_EG, is_IS, nb_NO
                value = value.replace("%b", str(date_val[1]))
                # some locales have %a for the abbreviated day, e.g. is_IS
                value = value.replace("%a", self._get_weekday(date_val))
                if date_val[0] == 0:  # ignore the zero day and its delimiter
                    i_day = value.find("%e")  # Icelandic uses %e and not %d
                    value = value.replace(value[i_day : i_day + 3], "")
                value = value.replace("%e", str(date_val[0]))
                value = value.replace("%Y", str(abs(date_val[2])))
                return value.replace("-", "/")


# -------------------------------------------------------------------------
#
# Register classes
#
# -------------------------------------------------------------------------
register_datehandler(
    ("is_IS", "is", "íslenskt", "Icelandic", ("%a %e.%b %Y",)),
    DateParserIs,
    DateDisplayIs,
)



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2007-2009  Gary Burton <gary.burton@zen.co.uk>
# Copyright (C) 2007-2009  Stephane Charette <stephanecharette@gmail.com>
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2008       Jason M. Simanek <jason@bohemianalps.com>
# Copyright (C) 2008-2011  Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010-      Serge Noiraud
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2016       Allen Crider
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

Classe:
    AddressBookPage
"""
# ------------------------------------------------
# python modules
# ------------------------------------------------
from decimal import getcontext
import logging

# ------------------------------------------------
# Gramps module
# ------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.plug.report import Bibliography
from gramps.plugins.lib.libhtml import Html

# ------------------------------------------------
# specific narrative web import
# ------------------------------------------------
from gramps.plugins.webreport.basepage import BasePage
from gramps.plugins.webreport.common import FULLCLEAR

_ = glocale.translation.sgettext
LOG = logging.getLogger(".NarrativeWeb")
getcontext().prec = 8


class AddressBookPage(BasePage):
    """
    Create one page for one Address
    """

    def __init__(
        self, report, the_lang, the_title, person_handle, has_add, has_res, has_url
    ):
        """
        @param: report        -- The instance of the main report class
                                 for this report
        @param: the_lang      -- The lang to process
        @param: the_title     -- The title page related to the language
        @param: person_handle -- the url, address and residence to use
                                 for the report
        @param: has_add       -- the address to use for the report
        @param: has_res       -- the residence to use for the report
        @param: has_url       -- the url to use for the report
        """
        person = report.database.get_person_from_handle(person_handle)
        BasePage.__init__(self, report, the_lang, the_title, person.gramps_id)
        self.bibli = Bibliography()

        self.uplink = True

        # set the file name and open file
        output_file, sio = self.report.create_file(person_handle, "addr")
        result = self.write_header(_("Address Book"))
        addressbookpage, dummy_head, dummy_body, outerwrapper = result

        # begin address book page division and section title
        with Html("div", class_="content", id="AddressBookDetail") as addressbookdetail:
            outerwrapper += addressbookdetail

            link = self.new_person_link(person_handle, uplink=True, person=person)
            addressbookdetail += Html("h3", link)

            # individual has an address
            if has_add:
                addressbookdetail += self.display_addr_list(has_add, None)

            # individual has a residence
            if has_res:
                addressbookdetail.extend(self.dump_residence(res) for res in has_res)

            # individual has a url
            if has_url:
                addressbookdetail += self.display_url_list(has_url)

        # add fullclear for proper styling
        # and footer section to page
        footer = self.write_footer(None)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the file
        self.xhtml_writer(addressbookpage, output_file, sio, 0)
